from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    addrwidth = 10
    myaxi = vthread.AXIM(m, 'myaxi', clk, rst, datawidth)
    ram_a = vthread.RAM(m, 'ram_a', clk, rst, datawidth, addrwidth)
    ram_b = vthread.RAM(m, 'ram_b', clk, rst, datawidth, addrwidth)

    strm = vthread.Stream(m, 'mystream', clk, rst)
    img_width = strm.constant('img_width')

    counter = strm.Counter()

    a = strm.source('a')
    a_addr = strm.Counter() - 1

    sp = strm.Scratchpad(a, a_addr, length=128)

    a_old_addr = strm.Counter() - img_width - 1
    a_old = sp.read(a_old_addr)

    b = a + a_old

    strm.sink(b, 'b', when=counter > img_width)

    def comp_stream(size, offset):
        strm.set_source('a', ram_a, offset, size * 2)
        strm.set_sink('b', ram_b, offset, size)
        strm.set_constant('img_width', size)
        strm.run()
        strm.join()

    def comp_sequential(size, offset):
        for i in range(size):
            a_buf = ram_a.read(i + offset)
            a = ram_a.read(i + offset + size)
            b = a_buf + a
            ram_b.write(i + offset, b)

    def check(size, offset_stream, offset_seq):
        all_ok = True
        for i in range(size):
            st = ram_b.read(i + offset_stream)
            sq = ram_b.read(i + offset_seq)
            if vthread.verilog.NotEql(st, sq):
                all_ok = False
        if all_ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

    def comp(size):
        # stream
        offset = 0
        myaxi.dma_read(ram_a, offset, 0, size * 2)
        comp_stream(size, offset)
        myaxi.dma_write(ram_b, offset, 1024, size)

        # sequential
        offset = size * 4
        myaxi.dma_read(ram_a, offset, 0, size * 2)
        comp_sequential(size, offset)
        myaxi.dma_write(ram_b, offset, 1024 * 2, size)

        # verification
        check(size, 0, offset)

        vthread.finish()

    th = vthread.Thread(m, 'th_comp', clk, rst, comp)
    fsm = th.start(32)

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst, memimg_name=memimg_name)
    memory.connect(ports, 'myaxi')

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    #simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function

from . import stypes
from .visitor import _Visitor


class _Scheduler(_Visitor):

    def max_stage(self, *vars):
        return stypes._max(*vars)

    def next_stage(self, node, stage):
        if stage is None:
            return 0
        return stage + node.latency

    def schedule(self, nodes):
        raise NotImplementedError()


class ASAPScheduler(_Scheduler):
    """ Determine the scheduled cycle and insert delay variables to fill the gap """

    def schedule(self, nodes):
        for node in sorted(nodes, key=lambda x: x.object_id):
            self.visit(node)

    def balance_output(self, nodes, max_stage):
        ret = []
        for node in sorted(nodes, key=lambda x: x.object_id):
            if not node._has_output():
                continue
            r = self.fill_gap(node, max_stage)
            t_data = node.output_data
            t_sig_data = node.output_sig_data
            node._disable_output()
            node._disable_output_sig()
            node._set_output_node(r)
            r._disable_output()
            r._disable_output_sig()
            r.output_data = t_data
            r.output_sig_data = t_sig_data
            ret.append(r)
        return ret

    def fill_gap(self, node, end_stage):
        if end_stage is None:
            return node
        if node.end_stage is None:
            return node
        if node.end_stage == end_stage:
            return node
        if node.end_stage > end_stage:
            raise ValueError("Illegal stage number: node.end_stage (%d) > end_stage (%d)" %
                             node.end_stage, end_stage)

        if isinstance(node, stypes._Delay) and node._get_parent_value() is not None:
            node = node._get_parent_value()

        prev = node
        cur_end_stage = prev.end_stage

        for i in range(cur_end_stage, end_stage):
            r = node._get_delayed_value(i + 1)
            if r is not None:
                prev = r
                cur_end_stage += 1
                continue
            r = stypes._Delay(prev)
            r._set_start_stage(cur_end_stage)
            r._set_end_stage(cur_end_stage + 1)
            node._add_delayed_value(i + 1, r)
            r._set_parent_value(node)
            prev = r
            cur_end_stage += 1

        return prev

    def visit__BinaryOperator(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        left = self.visit(node.left)
        right = self.visit(node.right)
        mine = self.max_stage(left, right)
        node.left = self.fill_gap(node.left, mine)
        node.right = self.fill_gap(node.right, mine)
        node._set_start_stage(mine)
        if getattr(node, 'variable_latency', None):
            node.latency = getattr(node, node.variable_latency)()
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit__UnaryOperator(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        right = self.visit(node.right)
        mine = self.max_stage(right)
        node.right = self.fill_gap(node.right, mine)
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit__SpecialOperator(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        ret = []
        for var in node.args:
            var = self.visit(var)
            ret.append(var)
        mine = self.max_stage(*ret)
        node.args = [self.fill_gap(var, mine) for var in node.args]
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit__Accumulator(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        right = self.visit(node.right)
        initval = self.visit(node.initval)
        enable = self.visit(node.enable) if node.enable is not None else None
        reset = self.visit(node.reset) if node.reset is not None else None
        mine = self.max_stage(right, initval, enable, reset)
        node.right = self.fill_gap(node.right, mine)
        node.initval = self.fill_gap(node.initval, mine)
        if node.enable is not None:
            node.enable = self.fill_gap(node.enable, mine)
        if node.reset is not None:
            node.reset = self.fill_gap(node.reset, mine)
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit_Substream(self, node):
        return self.visit__SpecialOperator(node)

    def visit_RingBuffer(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        right = self.visit(node.right)
        enable = self.visit(node.enable) if node.enable is not None else None
        reset = self.visit(node.reset) if node.reset is not None else None
        mine = self.max_stage(right, enable, reset)
        node.right = self.fill_gap(node.right, mine)
        if node.enable is not None:
            node.enable = self.fill_gap(node.enable, mine)
        if node.reset is not None:
            node.reset = self.fill_gap(node.reset, mine)
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit__RingBufferOutput(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        left = self.visit(node.left)
        right = self.visit(node.right)
        enable = self.visit(node.enable) if node.enable is not None else None
        reset = self.visit(node.reset) if node.reset is not None else None
        mine = self.max_stage(left, right, enable, reset)
        node.left = self.fill_gap(node.left, mine)
        node.right = self.fill_gap(node.right, mine)
        if node.enable is not None:
            node.enable = self.fill_gap(node.enable, mine)
        if node.reset is not None:
            node.reset = self.fill_gap(node.reset, mine)
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit_Scratchpad(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        left = self.visit(node.left)
        right = self.visit(node.right)
        enable = self.visit(node.enable) if node.enable is not None else None
        reset = self.visit(node.reset) if node.reset is not None else None
        mine = self.max_stage(left, right, enable, reset)
        node.left = self.fill_gap(node.left, mine)
        node.right = self.fill_gap(node.right, mine)
        if node.enable is not None:
            node.enable = self.fill_gap(node.enable, mine)
        if node.reset is not None:
            node.reset = self.fill_gap(node.reset, mine)
        node._set_start_stage(mine)
        end = self.next_stage(node, mine)
        node._set_end_stage(end)
        return end

    def visit__ScratchpadOutput(self, node):
        return self.visit__BinaryOperator(node)

    def visit__ParameterVariable(self, node):
        return None

    def visit__Variable(self, node):
        if node._has_start_stage():
            return node._get_end_stage()
        if isinstance(node.input_data, stypes._Numeric):
            data = self.visit(node.input_data)
            node._set_start_stage(data)
            return data
        mine = 0
        node._set_start_stage(mine)
        end = mine
        node._set_end_stage(end)
        return end

    def visit__Constant(self, node):
        return None



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import primitive_mux

expected_verilog = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED
);

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      count <= _tmp_3;
    end
  end

  wire _tmp_0;
  wire [32-1:0] _tmp_1;
  wire [32-1:0] _tmp_2;
  wire [32-1:0] _tmp_3;
  assign _tmp_0 = count == 1023;
  assign _tmp_1 = 0;
  assign _tmp_2 = count + 1;

  prim_mux
  #(
    .WIDTH(32)
  )
  inst_prim_mux__tmp_3
  (
    .sel(_tmp_0),
    .ina(_tmp_1),
    .inb(_tmp_2),
    .out(_tmp_3)
  );


  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      LED <= _tmp_7;
    end
  end

  wire _tmp_4;
  wire [WIDTH-1:0] _tmp_5;
  wire [WIDTH-1:0] _tmp_6;
  wire [WIDTH-1:0] _tmp_7;
  assign _tmp_4 = count == 1023;
  assign _tmp_5 = LED + 1;
  assign _tmp_6 = LED;

  prim_mux
  #(
    .WIDTH(WIDTH)
  )
  inst_prim_mux__tmp_7
  (
    .sel(_tmp_4),
    .ina(_tmp_5),
    .inb(_tmp_6),
    .out(_tmp_7)
  );


endmodule



module prim_mux #
(
  parameter WIDTH = 1
)
(
  input sel,
  input [WIDTH-1:0] ina,
  input [WIDTH-1:0] inb,
  output [WIDTH-1:0] out
);

  assign out = (sel)? ina : inb;

endmodule
"""

def test():
    veriloggen.reset()
    test_module = primitive_mux.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_axi_write

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [32-1:0] myaxi_awaddr;
  wire [8-1:0] myaxi_awlen;
  wire [3-1:0] myaxi_awsize;
  wire [2-1:0] myaxi_awburst;
  wire [2-1:0] myaxi_awlock;
  wire [4-1:0] myaxi_awcache;
  wire [3-1:0] myaxi_awprot;
  wire [4-1:0] myaxi_awqos;
  wire [2-1:0] myaxi_awuser;
  wire myaxi_awvalid;
  reg myaxi_awready;
  wire [32-1:0] myaxi_wdata;
  wire [4-1:0] myaxi_wstrb;
  wire myaxi_wlast;
  wire myaxi_wvalid;
  reg myaxi_wready;
  reg [2-1:0] myaxi_bresp;
  reg myaxi_bvalid;
  wire myaxi_bready;
  wire [32-1:0] myaxi_araddr;
  wire [8-1:0] myaxi_arlen;
  wire [3-1:0] myaxi_arsize;
  wire [2-1:0] myaxi_arburst;
  wire [2-1:0] myaxi_arlock;
  wire [4-1:0] myaxi_arcache;
  wire [3-1:0] myaxi_arprot;
  wire [4-1:0] myaxi_arqos;
  wire [2-1:0] myaxi_aruser;
  wire myaxi_arvalid;
  reg myaxi_arready;
  reg [32-1:0] myaxi_rdata;
  reg [2-1:0] myaxi_rresp;
  reg myaxi_rlast;
  reg myaxi_rvalid;
  wire myaxi_rready;
  reg [32-1:0] waddr;
  localparam waddr_init = 0;
  reg [32-1:0] _awlen;
  wire _tmp_0;
  assign _tmp_0 = 0;

  always @(*) begin
    myaxi_arready <= _tmp_0;
  end

  wire _tmp_1;
  assign _tmp_1 = 0;

  always @(*) begin
    myaxi_rvalid <= _tmp_1;
  end

  wire [32-1:0] _tmp_2;
  assign _tmp_2 = 0;

  always @(*) begin
    myaxi_rdata <= _tmp_2;
  end

  wire _tmp_3;
  assign _tmp_3 = 0;

  always @(*) begin
    myaxi_rlast <= _tmp_3;
  end


  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .myaxi_awaddr(myaxi_awaddr),
    .myaxi_awlen(myaxi_awlen),
    .myaxi_awsize(myaxi_awsize),
    .myaxi_awburst(myaxi_awburst),
    .myaxi_awlock(myaxi_awlock),
    .myaxi_awcache(myaxi_awcache),
    .myaxi_awprot(myaxi_awprot),
    .myaxi_awqos(myaxi_awqos),
    .myaxi_awuser(myaxi_awuser),
    .myaxi_awvalid(myaxi_awvalid),
    .myaxi_awready(myaxi_awready),
    .myaxi_wdata(myaxi_wdata),
    .myaxi_wstrb(myaxi_wstrb),
    .myaxi_wlast(myaxi_wlast),
    .myaxi_wvalid(myaxi_wvalid),
    .myaxi_wready(myaxi_wready),
    .myaxi_bresp(myaxi_bresp),
    .myaxi_bvalid(myaxi_bvalid),
    .myaxi_bready(myaxi_bready),
    .myaxi_araddr(myaxi_araddr),
    .myaxi_arlen(myaxi_arlen),
    .myaxi_arsize(myaxi_arsize),
    .myaxi_arburst(myaxi_arburst),
    .myaxi_arlock(myaxi_arlock),
    .myaxi_arcache(myaxi_arcache),
    .myaxi_arprot(myaxi_arprot),
    .myaxi_arqos(myaxi_arqos),
    .myaxi_aruser(myaxi_aruser),
    .myaxi_arvalid(myaxi_arvalid),
    .myaxi_arready(myaxi_arready),
    .myaxi_rdata(myaxi_rdata),
    .myaxi_rresp(myaxi_rresp),
    .myaxi_rlast(myaxi_rlast),
    .myaxi_rvalid(myaxi_rvalid),
    .myaxi_rready(myaxi_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, myaxi_awaddr, myaxi_awlen, myaxi_awsize, myaxi_awburst, myaxi_awlock, myaxi_awcache, myaxi_awprot, myaxi_awqos, myaxi_awuser, myaxi_awvalid, myaxi_awready, myaxi_wdata, myaxi_wstrb, myaxi_wlast, myaxi_wvalid, myaxi_wready, myaxi_bresp, myaxi_bvalid, myaxi_bready, myaxi_araddr, myaxi_arlen, myaxi_arsize, myaxi_arburst, myaxi_arlock, myaxi_arcache, myaxi_arprot, myaxi_arqos, myaxi_aruser, myaxi_arvalid, myaxi_arready, myaxi_rdata, myaxi_rresp, myaxi_rlast, myaxi_rvalid, myaxi_rready, waddr, _awlen, _tmp_0, _tmp_1, _tmp_2, _tmp_3);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    waddr = waddr_init;
    _awlen = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end

  localparam waddr_1 = 1;
  localparam waddr_2 = 2;
  localparam waddr_3 = 3;
  localparam waddr_4 = 4;
  localparam waddr_5 = 5;

  always @(posedge CLK) begin
    if(RST) begin
      waddr <= waddr_init;
      _awlen <= 0;
    end else begin
      case(waddr)
        waddr_init: begin
          myaxi_awready <= 0;
          myaxi_wready <= 0;
          _awlen <= 0;
          if(myaxi_awvalid) begin
            waddr <= waddr_1;
          end 
        end
        waddr_1: begin
          if(myaxi_awvalid) begin
            myaxi_awready <= 1;
          end 
          waddr <= waddr_2;
        end
        waddr_2: begin
          myaxi_awready <= 0;
          _awlen <= myaxi_awlen;
          waddr <= waddr_3;
        end
        waddr_3: begin
          myaxi_wready <= 0;
          if(myaxi_wvalid) begin
            waddr <= waddr_4;
          end 
        end
        waddr_4: begin
          if(myaxi_wvalid) begin
            myaxi_wready <= 1;
          end 
          waddr <= waddr_5;
        end
        waddr_5: begin
          myaxi_wready <= 0;
          _awlen <= _awlen - 1;
          waddr <= waddr_3;
          if(_awlen == 0) begin
            waddr <= waddr_init;
          end 
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  output reg [32-1:0] myaxi_awaddr,
  output reg [8-1:0] myaxi_awlen,
  output [3-1:0] myaxi_awsize,
  output [2-1:0] myaxi_awburst,
  output [2-1:0] myaxi_awlock,
  output [4-1:0] myaxi_awcache,
  output [3-1:0] myaxi_awprot,
  output [4-1:0] myaxi_awqos,
  output [2-1:0] myaxi_awuser,
  output reg myaxi_awvalid,
  input myaxi_awready,
  output reg [32-1:0] myaxi_wdata,
  output reg [4-1:0] myaxi_wstrb,
  output reg myaxi_wlast,
  output reg myaxi_wvalid,
  input myaxi_wready,
  input [2-1:0] myaxi_bresp,
  input myaxi_bvalid,
  output myaxi_bready,
  output reg [32-1:0] myaxi_araddr,
  output reg [8-1:0] myaxi_arlen,
  output [3-1:0] myaxi_arsize,
  output [2-1:0] myaxi_arburst,
  output [2-1:0] myaxi_arlock,
  output [4-1:0] myaxi_arcache,
  output [3-1:0] myaxi_arprot,
  output [4-1:0] myaxi_arqos,
  output [2-1:0] myaxi_aruser,
  output reg myaxi_arvalid,
  input myaxi_arready,
  input [32-1:0] myaxi_rdata,
  input [2-1:0] myaxi_rresp,
  input myaxi_rlast,
  input myaxi_rvalid,
  output myaxi_rready
);

  assign myaxi_awsize = 2;
  assign myaxi_awburst = 1;
  assign myaxi_awlock = 0;
  assign myaxi_awcache = 3;
  assign myaxi_awprot = 0;
  assign myaxi_awqos = 0;
  assign myaxi_awuser = 0;
  assign myaxi_bready = 1;
  assign myaxi_arsize = 2;
  assign myaxi_arburst = 1;
  assign myaxi_arlock = 0;
  assign myaxi_arcache = 3;
  assign myaxi_arprot = 0;
  assign myaxi_arqos = 0;
  assign myaxi_aruser = 0;
  assign myaxi_rready = 0;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg _myaxi_cond_0_1;
  reg [32-1:0] wdata;
  reg _tmp_1;
  reg _myaxi_cond_1_1;
  reg [32-1:0] sum;
  reg _seq_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      myaxi_araddr <= 0;
      myaxi_arlen <= 0;
      myaxi_arvalid <= 0;
      myaxi_awaddr <= 0;
      myaxi_awlen <= 0;
      myaxi_awvalid <= 0;
      _tmp_0 <= 0;
      _myaxi_cond_0_1 <= 0;
      myaxi_wdata <= 0;
      myaxi_wvalid <= 0;
      myaxi_wlast <= 0;
      myaxi_wstrb <= 0;
      _tmp_1 <= 0;
      _myaxi_cond_1_1 <= 0;
    end else begin
      if(_myaxi_cond_0_1) begin
        myaxi_awvalid <= 0;
      end 
      if(_myaxi_cond_1_1) begin
        myaxi_wvalid <= 0;
        myaxi_wlast <= 0;
        _tmp_1 <= 0;
      end 
      myaxi_araddr <= 0;
      myaxi_arlen <= 0;
      myaxi_arvalid <= 0;
      if((fsm == 0) && ((myaxi_awready || !myaxi_awvalid) && (_tmp_0 == 0))) begin
        myaxi_awaddr <= 1024;
        myaxi_awlen <= 63;
        myaxi_awvalid <= 1;
        _tmp_0 <= 64;
      end 
      if((fsm == 0) && ((myaxi_awready || !myaxi_awvalid) && (_tmp_0 == 0)) && 0) begin
        myaxi_awvalid <= 0;
      end 
      _myaxi_cond_0_1 <= 1;
      if(myaxi_awvalid && !myaxi_awready) begin
        myaxi_awvalid <= myaxi_awvalid;
      end 
      if((fsm == 1) && ((_tmp_0 > 0) && (myaxi_wready || !myaxi_wvalid) && (_tmp_0 > 0))) begin
        myaxi_wdata <= wdata;
        myaxi_wvalid <= 1;
        myaxi_wlast <= 0;
        myaxi_wstrb <= { 4{ 1'd1 } };
        _tmp_0 <= _tmp_0 - 1;
      end 
      if((fsm == 1) && ((_tmp_0 > 0) && (myaxi_wready || !myaxi_wvalid) && (_tmp_0 > 0)) && (_tmp_0 == 1)) begin
        myaxi_wlast <= 1;
        _tmp_1 <= 1;
      end 
      _myaxi_cond_1_1 <= 1;
      if(myaxi_wvalid && !myaxi_wready) begin
        myaxi_wvalid <= myaxi_wvalid;
        myaxi_wlast <= myaxi_wlast;
        _tmp_1 <= _tmp_1;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      wdata <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(myaxi_awready || !myaxi_awvalid) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if((_tmp_0 > 0) && (myaxi_wready || !myaxi_wvalid)) begin
            wdata <= wdata + 1;
          end 
          if(_tmp_1) begin
            fsm <= fsm_2;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      sum <= 0;
      _seq_cond_0_1 <= 0;
    end else begin
      if(_seq_cond_0_1) begin
        $display("sum=%d expected_sum=%d", sum, 2016);
      end 
      if(myaxi_wvalid && myaxi_wready) begin
        sum <= sum + myaxi_wdata;
      end 
      _seq_cond_0_1 <= myaxi_wvalid && myaxi_wready && myaxi_wlast;
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_axi_write.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import copy

import veriloggen.core.vtypes as vtypes


class CollectVisitor(object):

    def __init__(self):
        self.names = set()

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        if node is None:
            return None
        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)
        if isinstance(node, vtypes._BinaryOperator):
            return self.visit__BinaryOperator(node)
        if isinstance(node, vtypes._UnaryOperator):
            return self.visit__UnaryOperator(node)

        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)

    def visit_Int(self, node):
        return

    def visit_Float(self, node):
        return

    def visit_Str(self, node):
        return

    def visit_bool(self, node):
        return

    def visit_int(self, node):
        return

    def visit_str(self, node):
        return

    def visit_float(self, node):
        return

    def visit__Variable(self, node):
        self.names.add(node.name)

        self.visit(node.width)
        if hasattr(node, 'initval'):
            self.visit(node.initval)

    def visit_Pointer(self, node):
        self.visit(node.var)
        self.visit(node.pos)

    def visit_Slice(self, node):
        self.visit(node.var)
        self.visit(node.msb)
        self.visit(node.lsb)

    def visit_Cat(self, node):
        for var in node.vars:
            self.visit(var)

    def visit_Repeat(self, node):
        self.visit(node.var)
        self.visit(node.times)

    def visit_Cond(self, node):
        self.visit(node.condition)
        self.visit(node.true_value)
        self.visit(node.false_value)

    def visit__BinaryOperator(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit__UnaryOperator(self, node):
        self.visit(node.right)

    def visit_EmbeddedNumeric(self, node):
        return



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_fixed_add

expected_verilog = """
module test;

  reg uut_CLK;
  reg uut_RST;

  blinkled
  uut
  (
    .CLK(uut_CLK),
    .RST(uut_RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, uut_CLK, uut_RST);
  end


  initial begin
    uut_CLK = 0;
    forever begin
      #5 uut_CLK = !uut_CLK;
    end
  end


  initial begin
    uut_RST = 0;
    #100;
    uut_RST = 1;
    #100;
    uut_RST = 0;
    #100000;
    $finish;
  end


endmodule



module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST
);

  reg signed [16-1:0] a;
  reg signed [16-1:0] b;
  reg signed [32-1:0] c;
  reg signed [32-1:0] d;
  reg signed [32-1:0] e;

  always @(posedge CLK) begin
    if(RST) begin
      a <= 64;
      b <= 'sd1;
      c <= -2048;
      d <= 'sd1;
      e <= 0;
    end else begin
      a <= b - 'sd256 >>> 4;
      b <= a - 'sd16 << 4;
      c <= b[15:0] + 1;
      d <= c + 'sd256;
      e <= e + 'sd65536;
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = types_fixed_add.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    count = m.Reg('count', 32)

    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            If(count == 1023)(
                count(0)
            ).Else(
                count(count + 1)
            )
        ))
    
    m.Always(Posedge(clk))(
        If(rst)(
            led(0)
        ).Else(
            If(count == 1024 - 1)(
                led(led + 1)
            )
        ))
    
    return m

def mkTest():
    m = Module('test')
    
    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)
    
    clk = ports['CLK']
    rst = ports['RST']
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))
    
    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)
    
    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    count = m.Reg('count', 32)

    patterns_count = []
    patterns_count.append( (count < 10, count + 1) )
    patterns_count.append( (count == 1023, 0) )
    patterns_count.append( (None, count + 1) )

    patterns_led = []
    patterns_led.append( (count < 10, led) )
    patterns_led.append( (count == 1023, led + 1) )
    patterns_led.append( (None, led) )

    m.Always(Posedge(clk))(
        If(rst)(
            count(0),
            led(0)
        ).Else(
            count(PatternMux(*patterns_count)),
            led(PatternMux(*patterns_led)),
        ))

    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_ipxact_master

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [32-1:0] LED;
  wire [32-1:0] myaxi_awaddr;
  wire [8-1:0] myaxi_awlen;
  wire [3-1:0] myaxi_awsize;
  wire [2-1:0] myaxi_awburst;
  wire [2-1:0] myaxi_awlock;
  wire [4-1:0] myaxi_awcache;
  wire [3-1:0] myaxi_awprot;
  wire [4-1:0] myaxi_awqos;
  wire [2-1:0] myaxi_awuser;
  wire myaxi_awvalid;
  reg myaxi_awready;
  wire [32-1:0] myaxi_wdata;
  wire [4-1:0] myaxi_wstrb;
  wire myaxi_wlast;
  wire myaxi_wvalid;
  reg myaxi_wready;
  reg [2-1:0] myaxi_bresp;
  reg myaxi_bvalid;
  wire myaxi_bready;
  wire [32-1:0] myaxi_araddr;
  wire [8-1:0] myaxi_arlen;
  wire [3-1:0] myaxi_arsize;
  wire [2-1:0] myaxi_arburst;
  wire [2-1:0] myaxi_arlock;
  wire [4-1:0] myaxi_arcache;
  wire [3-1:0] myaxi_arprot;
  wire [4-1:0] myaxi_arqos;
  wire [2-1:0] myaxi_aruser;
  wire myaxi_arvalid;
  reg myaxi_arready;
  reg [32-1:0] myaxi_rdata;
  reg [2-1:0] myaxi_rresp;
  reg myaxi_rlast;
  reg myaxi_rvalid;
  wire myaxi_rready;
  wire [32-1:0] memory_awaddr;
  wire [8-1:0] memory_awlen;
  wire [3-1:0] memory_awsize;
  wire [2-1:0] memory_awburst;
  wire [2-1:0] memory_awlock;
  wire [4-1:0] memory_awcache;
  wire [3-1:0] memory_awprot;
  wire [4-1:0] memory_awqos;
  wire [2-1:0] memory_awuser;
  wire memory_awvalid;
  reg memory_awready;
  wire [32-1:0] memory_wdata;
  wire [4-1:0] memory_wstrb;
  wire memory_wlast;
  wire memory_wvalid;
  reg memory_wready;
  wire [2-1:0] memory_bresp;
  reg memory_bvalid;
  wire memory_bready;
  wire [32-1:0] memory_araddr;
  wire [8-1:0] memory_arlen;
  wire [3-1:0] memory_arsize;
  wire [2-1:0] memory_arburst;
  wire [2-1:0] memory_arlock;
  wire [4-1:0] memory_arcache;
  wire [3-1:0] memory_arprot;
  wire [4-1:0] memory_arqos;
  wire [2-1:0] memory_aruser;
  wire memory_arvalid;
  reg memory_arready;
  reg [32-1:0] memory_rdata;
  wire [2-1:0] memory_rresp;
  reg memory_rlast;
  reg memory_rvalid;
  wire memory_rready;
  assign memory_bresp = 0;
  assign memory_rresp = 0;
  reg [32-1:0] _memory_fsm;
  localparam _memory_fsm_init = 0;
  reg [8-1:0] _memory_mem [0:2**20-1];

  initial begin
    $readmemh("_memory_memimg_.out", _memory_mem);
  end

  reg [33-1:0] _write_count;
  reg [32-1:0] _write_addr;
  reg [33-1:0] _read_count;
  reg [32-1:0] _read_addr;
  reg [33-1:0] _sleep_count;
  reg [32-1:0] _d1__memory_fsm;
  reg __memory_fsm_cond_100_0_1;
  reg __memory_fsm_cond_200_1_1;
  reg __memory_fsm_cond_211_2_1;
  assign memory_awaddr = myaxi_awaddr;
  assign memory_awlen = myaxi_awlen;
  assign memory_awsize = myaxi_awsize;
  assign memory_awburst = myaxi_awburst;
  assign memory_awlock = myaxi_awlock;
  assign memory_awcache = myaxi_awcache;
  assign memory_awprot = myaxi_awprot;
  assign memory_awqos = myaxi_awqos;
  assign memory_awuser = myaxi_awuser;
  assign memory_awvalid = myaxi_awvalid;
  wire _tmp_0;
  assign _tmp_0 = memory_awready;

  always @(*) begin
    myaxi_awready = _tmp_0;
  end

  assign memory_wdata = myaxi_wdata;
  assign memory_wstrb = myaxi_wstrb;
  assign memory_wlast = myaxi_wlast;
  assign memory_wvalid = myaxi_wvalid;
  wire _tmp_1;
  assign _tmp_1 = memory_wready;

  always @(*) begin
    myaxi_wready = _tmp_1;
  end

  wire [2-1:0] _tmp_2;
  assign _tmp_2 = memory_bresp;

  always @(*) begin
    myaxi_bresp = _tmp_2;
  end

  wire _tmp_3;
  assign _tmp_3 = memory_bvalid;

  always @(*) begin
    myaxi_bvalid = _tmp_3;
  end

  assign memory_bready = myaxi_bready;
  assign memory_araddr = myaxi_araddr;
  assign memory_arlen = myaxi_arlen;
  assign memory_arsize = myaxi_arsize;
  assign memory_arburst = myaxi_arburst;
  assign memory_arlock = myaxi_arlock;
  assign memory_arcache = myaxi_arcache;
  assign memory_arprot = myaxi_arprot;
  assign memory_arqos = myaxi_arqos;
  assign memory_aruser = myaxi_aruser;
  assign memory_arvalid = myaxi_arvalid;
  wire _tmp_4;
  assign _tmp_4 = memory_arready;

  always @(*) begin
    myaxi_arready = _tmp_4;
  end

  wire [32-1:0] _tmp_5;
  assign _tmp_5 = memory_rdata;

  always @(*) begin
    myaxi_rdata = _tmp_5;
  end

  wire [2-1:0] _tmp_6;
  assign _tmp_6 = memory_rresp;

  always @(*) begin
    myaxi_rresp = _tmp_6;
  end

  wire _tmp_7;
  assign _tmp_7 = memory_rlast;

  always @(*) begin
    myaxi_rlast = _tmp_7;
  end

  wire _tmp_8;
  assign _tmp_8 = memory_rvalid;

  always @(*) begin
    myaxi_rvalid = _tmp_8;
  end

  assign memory_rready = myaxi_rready;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED),
    .myaxi_awaddr(myaxi_awaddr),
    .myaxi_awlen(myaxi_awlen),
    .myaxi_awsize(myaxi_awsize),
    .myaxi_awburst(myaxi_awburst),
    .myaxi_awlock(myaxi_awlock),
    .myaxi_awcache(myaxi_awcache),
    .myaxi_awprot(myaxi_awprot),
    .myaxi_awqos(myaxi_awqos),
    .myaxi_awuser(myaxi_awuser),
    .myaxi_awvalid(myaxi_awvalid),
    .myaxi_awready(myaxi_awready),
    .myaxi_wdata(myaxi_wdata),
    .myaxi_wstrb(myaxi_wstrb),
    .myaxi_wlast(myaxi_wlast),
    .myaxi_wvalid(myaxi_wvalid),
    .myaxi_wready(myaxi_wready),
    .myaxi_bresp(myaxi_bresp),
    .myaxi_bvalid(myaxi_bvalid),
    .myaxi_bready(myaxi_bready),
    .myaxi_araddr(myaxi_araddr),
    .myaxi_arlen(myaxi_arlen),
    .myaxi_arsize(myaxi_arsize),
    .myaxi_arburst(myaxi_arburst),
    .myaxi_arlock(myaxi_arlock),
    .myaxi_arcache(myaxi_arcache),
    .myaxi_arprot(myaxi_arprot),
    .myaxi_arqos(myaxi_arqos),
    .myaxi_aruser(myaxi_aruser),
    .myaxi_arvalid(myaxi_arvalid),
    .myaxi_arready(myaxi_arready),
    .myaxi_rdata(myaxi_rdata),
    .myaxi_rresp(myaxi_rresp),
    .myaxi_rlast(myaxi_rlast),
    .myaxi_rvalid(myaxi_rvalid),
    .myaxi_rready(myaxi_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, LED, myaxi_awaddr, myaxi_awlen, myaxi_awsize, myaxi_awburst, myaxi_awlock, myaxi_awcache, myaxi_awprot, myaxi_awqos, myaxi_awuser, myaxi_awvalid, myaxi_awready, myaxi_wdata, myaxi_wstrb, myaxi_wlast, myaxi_wvalid, myaxi_wready, myaxi_bresp, myaxi_bvalid, myaxi_bready, myaxi_araddr, myaxi_arlen, myaxi_arsize, myaxi_arburst, myaxi_arlock, myaxi_arcache, myaxi_arprot, myaxi_arqos, myaxi_aruser, myaxi_arvalid, myaxi_arready, myaxi_rdata, myaxi_rresp, myaxi_rlast, myaxi_rvalid, myaxi_rready, memory_awaddr, memory_awlen, memory_awsize, memory_awburst, memory_awlock, memory_awcache, memory_awprot, memory_awqos, memory_awuser, memory_awvalid, memory_awready, memory_wdata, memory_wstrb, memory_wlast, memory_wvalid, memory_wready, memory_bresp, memory_bvalid, memory_bready, memory_araddr, memory_arlen, memory_arsize, memory_arburst, memory_arlock, memory_arcache, memory_arprot, memory_arqos, memory_aruser, memory_arvalid, memory_arready, memory_rdata, memory_rresp, memory_rlast, memory_rvalid, memory_rready, _memory_fsm, _write_count, _write_addr, _read_count, _read_addr, _sleep_count, _d1__memory_fsm, __memory_fsm_cond_100_0_1, __memory_fsm_cond_200_1_1, __memory_fsm_cond_211_2_1, _tmp_0, _tmp_1, _tmp_2, _tmp_3, _tmp_4, _tmp_5, _tmp_6, _tmp_7, _tmp_8);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    memory_awready = 0;
    memory_wready = 0;
    memory_bvalid = 0;
    memory_arready = 0;
    memory_rdata = 0;
    memory_rlast = 0;
    memory_rvalid = 0;
    _memory_fsm = _memory_fsm_init;
    _write_count = 0;
    _write_addr = 0;
    _read_count = 0;
    _read_addr = 0;
    _sleep_count = 0;
    _d1__memory_fsm = _memory_fsm_init;
    __memory_fsm_cond_100_0_1 = 0;
    __memory_fsm_cond_200_1_1 = 0;
    __memory_fsm_cond_211_2_1 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end

  localparam _memory_fsm_200 = 200;
  localparam _memory_fsm_201 = 201;
  localparam _memory_fsm_202 = 202;
  localparam _memory_fsm_203 = 203;
  localparam _memory_fsm_204 = 204;
  localparam _memory_fsm_205 = 205;
  localparam _memory_fsm_206 = 206;
  localparam _memory_fsm_207 = 207;
  localparam _memory_fsm_208 = 208;
  localparam _memory_fsm_209 = 209;
  localparam _memory_fsm_210 = 210;
  localparam _memory_fsm_211 = 211;
  localparam _memory_fsm_100 = 100;
  localparam _memory_fsm_101 = 101;
  localparam _memory_fsm_102 = 102;
  localparam _memory_fsm_103 = 103;
  localparam _memory_fsm_104 = 104;
  localparam _memory_fsm_105 = 105;
  localparam _memory_fsm_106 = 106;
  localparam _memory_fsm_107 = 107;
  localparam _memory_fsm_108 = 108;
  localparam _memory_fsm_109 = 109;
  localparam _memory_fsm_110 = 110;
  localparam _memory_fsm_111 = 111;
  localparam _memory_fsm_112 = 112;

  always @(posedge CLK) begin
    if(RST) begin
      _memory_fsm <= _memory_fsm_init;
      _d1__memory_fsm <= _memory_fsm_init;
      memory_awready <= 0;
      _write_addr <= 0;
      _write_count <= 0;
      __memory_fsm_cond_100_0_1 <= 0;
      memory_wready <= 0;
      memory_arready <= 0;
      _read_addr <= 0;
      _read_count <= 0;
      __memory_fsm_cond_200_1_1 <= 0;
      memory_rdata[7:0] <= (0 >> 0) & { 8{ 1'd1 } };
      memory_rdata[15:8] <= (0 >> 8) & { 8{ 1'd1 } };
      memory_rdata[23:16] <= (0 >> 16) & { 8{ 1'd1 } };
      memory_rdata[31:24] <= (0 >> 24) & { 8{ 1'd1 } };
      memory_rvalid <= 0;
      memory_rlast <= 0;
      __memory_fsm_cond_211_2_1 <= 0;
      memory_rdata <= 0;
      memory_bvalid <= 0;
      _sleep_count <= 0;
    end else begin
      if(memory_bvalid && memory_bready) begin
        memory_bvalid <= 0;
      end 
      if(memory_wvalid && memory_wready && memory_wlast) begin
        memory_bvalid <= 1;
      end 
      _sleep_count <= _sleep_count + 1;
      if(_sleep_count == 3) begin
        _sleep_count <= 0;
      end 
      _d1__memory_fsm <= _memory_fsm;
      case(_d1__memory_fsm)
        _memory_fsm_100: begin
          if(__memory_fsm_cond_100_0_1) begin
            memory_awready <= 0;
          end 
        end
        _memory_fsm_200: begin
          if(__memory_fsm_cond_200_1_1) begin
            memory_arready <= 0;
          end 
        end
        _memory_fsm_211: begin
          if(__memory_fsm_cond_211_2_1) begin
            memory_rvalid <= 0;
            memory_rlast <= 0;
          end 
        end
      endcase
      case(_memory_fsm)
        _memory_fsm_init: begin
          if(memory_awvalid) begin
            _memory_fsm <= _memory_fsm_100;
          end 
          if(memory_arvalid) begin
            _memory_fsm <= _memory_fsm_200;
          end 
        end
        _memory_fsm_100: begin
          if(memory_awvalid && !memory_bvalid) begin
            memory_awready <= 1;
            _write_addr <= memory_awaddr;
            _write_count <= memory_awlen + 1;
          end 
          __memory_fsm_cond_100_0_1 <= 1;
          if(!memory_awvalid) begin
            _memory_fsm <= _memory_fsm_init;
          end 
          if(memory_awvalid) begin
            _memory_fsm <= _memory_fsm_101;
          end 
        end
        _memory_fsm_101: begin
          _memory_fsm <= _memory_fsm_102;
        end
        _memory_fsm_102: begin
          _memory_fsm <= _memory_fsm_103;
        end
        _memory_fsm_103: begin
          _memory_fsm <= _memory_fsm_104;
        end
        _memory_fsm_104: begin
          _memory_fsm <= _memory_fsm_105;
        end
        _memory_fsm_105: begin
          _memory_fsm <= _memory_fsm_106;
        end
        _memory_fsm_106: begin
          _memory_fsm <= _memory_fsm_107;
        end
        _memory_fsm_107: begin
          _memory_fsm <= _memory_fsm_108;
        end
        _memory_fsm_108: begin
          _memory_fsm <= _memory_fsm_109;
        end
        _memory_fsm_109: begin
          _memory_fsm <= _memory_fsm_110;
        end
        _memory_fsm_110: begin
          _memory_fsm <= _memory_fsm_111;
        end
        _memory_fsm_111: begin
          memory_wready <= 1;
          _memory_fsm <= _memory_fsm_112;
        end
        _memory_fsm_112: begin
          if(memory_wvalid && memory_wstrb[0]) begin
            _memory_mem[_write_addr + 0] <= memory_wdata[7:0];
          end 
          if(memory_wvalid && memory_wstrb[1]) begin
            _memory_mem[_write_addr + 1] <= memory_wdata[15:8];
          end 
          if(memory_wvalid && memory_wstrb[2]) begin
            _memory_mem[_write_addr + 2] <= memory_wdata[23:16];
          end 
          if(memory_wvalid && memory_wstrb[3]) begin
            _memory_mem[_write_addr + 3] <= memory_wdata[31:24];
          end 
          if(memory_wvalid && memory_wready) begin
            _write_addr <= _write_addr + 4;
            _write_count <= _write_count - 1;
          end 
          if(_sleep_count == 3) begin
            memory_wready <= 0;
          end else begin
            memory_wready <= 1;
          end
          if(memory_wvalid && memory_wready && (_write_count == 1)) begin
            memory_wready <= 0;
          end 
          if(memory_wvalid && memory_wready && (_write_count == 1)) begin
            _memory_fsm <= _memory_fsm_init;
          end 
        end
        _memory_fsm_200: begin
          if(memory_arvalid) begin
            memory_arready <= 1;
            _read_addr <= memory_araddr;
            _read_count <= memory_arlen + 1;
          end 
          __memory_fsm_cond_200_1_1 <= 1;
          if(!memory_arvalid) begin
            _memory_fsm <= _memory_fsm_init;
          end 
          if(memory_arvalid) begin
            _memory_fsm <= _memory_fsm_201;
          end 
        end
        _memory_fsm_201: begin
          _memory_fsm <= _memory_fsm_202;
        end
        _memory_fsm_202: begin
          _memory_fsm <= _memory_fsm_203;
        end
        _memory_fsm_203: begin
          _memory_fsm <= _memory_fsm_204;
        end
        _memory_fsm_204: begin
          _memory_fsm <= _memory_fsm_205;
        end
        _memory_fsm_205: begin
          _memory_fsm <= _memory_fsm_206;
        end
        _memory_fsm_206: begin
          _memory_fsm <= _memory_fsm_207;
        end
        _memory_fsm_207: begin
          _memory_fsm <= _memory_fsm_208;
        end
        _memory_fsm_208: begin
          _memory_fsm <= _memory_fsm_209;
        end
        _memory_fsm_209: begin
          _memory_fsm <= _memory_fsm_210;
        end
        _memory_fsm_210: begin
          _memory_fsm <= _memory_fsm_211;
        end
        _memory_fsm_211: begin
          if(memory_rready | !memory_rvalid) begin
            memory_rdata[7:0] <= _memory_mem[_read_addr + 0];
          end 
          if(memory_rready | !memory_rvalid) begin
            memory_rdata[15:8] <= _memory_mem[_read_addr + 1];
          end 
          if(memory_rready | !memory_rvalid) begin
            memory_rdata[23:16] <= _memory_mem[_read_addr + 2];
          end 
          if(memory_rready | !memory_rvalid) begin
            memory_rdata[31:24] <= _memory_mem[_read_addr + 3];
          end 
          if((_sleep_count < 3) && (_read_count > 0) && memory_rready | !memory_rvalid) begin
            memory_rvalid <= 1;
            _read_addr <= _read_addr + 4;
            _read_count <= _read_count - 1;
          end 
          if((_sleep_count < 3) && (_read_count == 1) && memory_rready | !memory_rvalid) begin
            memory_rlast <= 1;
          end 
          __memory_fsm_cond_211_2_1 <= 1;
          if(memory_rvalid && !memory_rready) begin
            memory_rvalid <= memory_rvalid;
            memory_rdata <= memory_rdata;
            memory_rlast <= memory_rlast;
          end 
          if(memory_rvalid && memory_rready && (_read_count == 0)) begin
            _memory_fsm <= _memory_fsm_init;
          end 
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  output [32-1:0] LED,
  output reg [32-1:0] myaxi_awaddr,
  output reg [8-1:0] myaxi_awlen,
  output [3-1:0] myaxi_awsize,
  output [2-1:0] myaxi_awburst,
  output [2-1:0] myaxi_awlock,
  output [4-1:0] myaxi_awcache,
  output [3-1:0] myaxi_awprot,
  output [4-1:0] myaxi_awqos,
  output [2-1:0] myaxi_awuser,
  output reg myaxi_awvalid,
  input myaxi_awready,
  output reg [32-1:0] myaxi_wdata,
  output reg [4-1:0] myaxi_wstrb,
  output reg myaxi_wlast,
  output reg myaxi_wvalid,
  input myaxi_wready,
  input [2-1:0] myaxi_bresp,
  input myaxi_bvalid,
  output myaxi_bready,
  output reg [32-1:0] myaxi_araddr,
  output reg [8-1:0] myaxi_arlen,
  output [3-1:0] myaxi_arsize,
  output [2-1:0] myaxi_arburst,
  output [2-1:0] myaxi_arlock,
  output [4-1:0] myaxi_arcache,
  output [3-1:0] myaxi_arprot,
  output [4-1:0] myaxi_arqos,
  output [2-1:0] myaxi_aruser,
  output reg myaxi_arvalid,
  input myaxi_arready,
  input [32-1:0] myaxi_rdata,
  input [2-1:0] myaxi_rresp,
  input myaxi_rlast,
  input myaxi_rvalid,
  output myaxi_rready
);

  assign myaxi_awsize = 2;
  assign myaxi_awburst = 1;
  assign myaxi_awlock = 0;
  assign myaxi_awcache = 3;
  assign myaxi_awprot = 0;
  assign myaxi_awqos = 0;
  assign myaxi_awuser = 0;
  assign myaxi_bready = 1;
  assign myaxi_arsize = 2;
  assign myaxi_arburst = 1;
  assign myaxi_arlock = 0;
  assign myaxi_arcache = 3;
  assign myaxi_arprot = 0;
  assign myaxi_arqos = 0;
  assign myaxi_aruser = 0;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] sum;
  assign LED = sum;
  reg [9-1:0] _tmp_0;
  reg _myaxi_cond_0_1;
  reg [9-1:0] _tmp_1;
  reg _myaxi_cond_1_1;
  assign myaxi_rready = (fsm == 1) || (fsm == 3);
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      sum <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(myaxi_arready || !myaxi_arvalid) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(myaxi_rready && myaxi_rvalid) begin
            sum <= sum + myaxi_rdata;
          end 
          if(myaxi_rready && myaxi_rvalid && myaxi_rlast) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          if(myaxi_arready || !myaxi_arvalid) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(myaxi_rready && myaxi_rvalid) begin
            sum <= sum + myaxi_rdata;
          end 
          if(myaxi_rready && myaxi_rvalid && myaxi_rlast) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          $display("sum=%d expected_sum=%d", sum, 53184);
          fsm <= fsm_5;
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      myaxi_awaddr <= 0;
      myaxi_awlen <= 0;
      myaxi_awvalid <= 0;
      myaxi_wdata <= 0;
      myaxi_wstrb <= 0;
      myaxi_wlast <= 0;
      myaxi_wvalid <= 0;
      myaxi_araddr <= 0;
      myaxi_arlen <= 0;
      myaxi_arvalid <= 0;
      _tmp_0 <= 0;
      _myaxi_cond_0_1 <= 0;
      _tmp_1 <= 0;
      _myaxi_cond_1_1 <= 0;
    end else begin
      if(_myaxi_cond_0_1) begin
        myaxi_arvalid <= 0;
      end 
      if(_myaxi_cond_1_1) begin
        myaxi_arvalid <= 0;
      end 
      myaxi_awaddr <= 0;
      myaxi_awlen <= 0;
      myaxi_awvalid <= 0;
      myaxi_wdata <= 0;
      myaxi_wstrb <= 0;
      myaxi_wlast <= 0;
      myaxi_wvalid <= 0;
      if((fsm == 0) && ((myaxi_arready || !myaxi_arvalid) && (_tmp_0 == 0))) begin
        myaxi_araddr <= 1024;
        myaxi_arlen <= 63;
        myaxi_arvalid <= 1;
        _tmp_0 <= 64;
      end 
      _myaxi_cond_0_1 <= 1;
      if(myaxi_arvalid && !myaxi_arready) begin
        myaxi_arvalid <= myaxi_arvalid;
      end 
      if(myaxi_rready && myaxi_rvalid && (_tmp_0 > 0)) begin
        _tmp_0 <= _tmp_0 - 1;
      end 
      if((fsm == 2) && ((myaxi_arready || !myaxi_arvalid) && (_tmp_1 == 0))) begin
        myaxi_araddr <= 2048;
        myaxi_arlen <= 63;
        myaxi_arvalid <= 1;
        _tmp_1 <= 64;
      end 
      _myaxi_cond_1_1 <= 1;
      if(myaxi_arvalid && !myaxi_arready) begin
        myaxi_arvalid <= myaxi_arvalid;
      end 
      if(myaxi_rready && myaxi_rvalid && (_tmp_1 > 0)) begin
        _tmp_1 <= _tmp_1 - 1;
      end 
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_ipxact_master.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    
    clk = m.Input('CLK')
    rst = m.Input('RST')

    x = m.Reg('x', 32, initval=0)
    y = m.Reg('y', 32, initval=0)
    z = m.Reg('z', 32, initval=0)
    
    seq_x = Seq(m, 'seq_x', clk, rst)
    seq_y = Seq(m, 'seq_y', clk, rst)
    seq_z = Seq(m, 'seq_z', clk, rst)
    
    seq_x.If(x < 16)(
        x.inc()
    )
    seq_x.If(x == 16).Delay(1)(
        x(0)
    )
    
    seq_y.If(y < 32)(
        y.inc()
    )
    seq_y.If(y == 32).Delay(2)(
        y(0)
    )
    
    prev_x = seq_x.Prev(x, 1)
    prev_y = seq_y.Prev(y, 1)
    
    seq_z(
        z(prev_x + prev_y)
    )
    
    seq_x.update(seq_y)
    seq_x.update(seq_z)

    seq_x.make_always()

    return m

def mkTest():
    m = Module('test')

    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    reset_done = m.Reg('reset_done', initval=0)
    reset_stmt = []
    reset_stmt.append( reset_done(0) )
    
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(10000),
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function

from . import stream

__intrinsics__ = ('copy', 'copy_pattern')

copy_stream_cache = {}
copy_pattern_stream_cache = {}


def copy(fsm, src_ram, dst_ram, src_addr, dst_addr, size,
         src_stride=1, dst_stride=1,
         src_port=0, dst_port=0):

    name = '_'.join(['copy', src_ram.name, dst_ram.name])

    if name in copy_stream_cache:
        strm = copy_stream_cache[name]

    else:
        m = src_ram.m
        clk = src_ram.clk
        rst = src_ram.rst
        strm = stream.Stream(m, name, clk, rst)
        src = strm.source('src')
        dst = src
        strm.sink(dst, 'dst')
        copy_stream_cache[name] = strm

    strm.set_source(fsm, 'src', src_ram, src_addr, size,
                    stride=src_stride, port=src_port)
    strm.set_sink(fsm, 'dst', dst_ram, dst_addr, size,
                  stride=dst_stride, port=dst_port)
    strm.run(fsm)
    strm.join(fsm)


def copy_pattern(fsm, src_ram, dst_ram, src_addr, dst_addr,
                 src_pattern, dst_pattern,
                 src_port=0, dst_port=0):

    name = '_'.join(['copy', src_ram.name, dst_ram.name])

    if name in copy_pattern_stream_cache:
        strm = copy_pattern_stream_cache[name]

    else:
        m = src_ram.m
        clk = src_ram.clk
        rst = src_ram.rst
        strm = stream.Stream(m, name, clk, rst)
        src = strm.source('src')
        dst = src
        strm.sink(dst, 'dst')
        copy_pattern_stream_cache[name] = strm

    strm.set_source_pattern(fsm, 'src', src_ram, src_addr, src_pattern,
                            port=src_port)
    strm.set_sink_pattern(fsm, 'dst', dst_ram, dst_addr, dst_pattern,
                          port=dst_port)
    strm.run(fsm)
    strm.join(fsm)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

class SeqIfElse(object):
    def __init__(self, condition, true_statements, false_statements=()):
        self.condition = condition
        self.true_statements = true_statements
        self.false_statements = false_statements

    def __len__(self):
        return len(self.true_statements) + len(self.true_statements)
        
def add_if_else(fsm, ifelse):
    # future index
    index_else = fsm.current + len(ifelse.true_statements) + 1
    index_merge = fsm.current + len(ifelse.true_statements) + len(ifelse.false_statements) + 1
    index_true = fsm.current + 1 if ifelse.true_statements else index_merge
    
    fsm.goto( dst=index_true, cond=ifelse.condition, else_dst=index_else ).inc()

    # then
    for i, s in enumerate(ifelse.true_statements):
        if i < len(ifelse.true_statements) - 1:
            fsm( *s )
            fsm.goto_next()
        else:
            fsm( *s )
            fsm.goto(index_merge).inc()

    # else
    for i, s in enumerate(ifelse.false_statements):
        if i < len(ifelse.false_statements) - 1:
            fsm( *s )
            fsm.goto_next()
        else:
            fsm( *s )
            fsm.goto_next()

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width, initval=0)
    count = m.Reg('count', 32, initval=0)

    fsm = FSM(m, 'fsm', clk, rst)
    # get the initial index (= 0)
    init = fsm.current

    fsm( count(count + 1) )
    fsm.goto_next()

    # if-then-else statements
    condition = count < 1024
    true_statements = [ [ count(count + 2) ],
                        [ count(count + 3) ] ]
    false_statements = [ [ led(led + 1) ],
                         [ led(led + 2) ] ]
    ifelse = SeqIfElse(condition, true_statements, false_statements)
    add_if_else(fsm, ifelse)
    
    # go to first
    fsm.goto(init)
    
    fsm.make_always()
    
    return m

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

try:
    log2 = math.log2
except:
    def log2(v):
        return math.log(v, 2)

def mkLed(numports=8):
    if log2(numports) % 1.0 != 0.0:
        raise ValueError('numports must be power of 2')
    
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    
    idata = [ m.Input('idata_%02d' % i, 32) for i in range(numports) ]
    ivalid = [ m.Input('ivalid_%02d' % i) for i in range(numports) ]
    odata = m.OutputReg('odata', 32, initval=0)
    ovalid = m.OutputReg('ovalid', initval=0)

    seq = Seq(m, 'seq', clk, rst)
    pdata = idata
    pvalid = ivalid
    ndata = []
    nvalid = []
    
    for s in range( int(log2(numports)) ):
        for i in range( numports >> (s+1) ):
            td = m.TmpReg(32, initval=0)
            tv = m.TmpReg(initval=0)
            ndata.append(td)
            nvalid.append(tv)
            cond = AndList(pvalid[i*2], pvalid[i*2+1])
            
            seq.If(cond)(
                td(pdata[i*2] + pdata[i*2+1])
            )
            seq(
                tv(cond)
            )
            
        pdata = ndata
        pvalid = nvalid
        ndata = []
        nvalid = []

    seq(
        odata(pdata[-1])
    )
    seq(
        ovalid(pvalid[-1])
    )
        
    seq.make_always()

    return m

def mkTest():
    m = Module('test')

    # target instance
    led = mkLed()
    
    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']
    idata = [ p for k, p in sorted(ports.items(), key=lambda x:x[0]) if k.startswith('idata') ]
    ivalid = [ p for k, p in sorted(ports.items(), key=lambda x:x[0]) if k.startswith('ivalid') ]
    
    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    reset_stmt = []
    for d in idata:
        reset_stmt.append( d(0) )
    for v in ivalid:
        reset_stmt.append( v(0) )
    
    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000),
        nclk(clk),
        
        [ d(0) for d in idata ],
        [ v(0) for v in ivalid ],
        nclk(clk),
        
        [ d(i+1) for i, d in enumerate(idata) ],
        [ v(1) for v in ivalid ],
        nclk(clk),
        
        [ d(1) for d in idata ],
        [ v(0) for v in ivalid ],
        nclk(clk),
        
        [ d(i+10) for i, d in enumerate(idata) ],
        [ v(1) for v in ivalid ],
        nclk(clk),
        
        [ v(0) for v in ivalid ],
        [ nclk(clk) for _ in range(10) ],
        
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_fixed_div

expected_verilog = """
module test;

  reg uut_CLK;
  reg uut_RST;

  blinkled
  uut
  (
    .CLK(uut_CLK),
    .RST(uut_RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, uut_CLK, uut_RST);
  end


  initial begin
    uut_CLK = 0;
    forever begin
      #5 uut_CLK = !uut_CLK;
    end
  end


  initial begin
    uut_RST = 0;
    #100;
    uut_RST = 1;
    #100;
    uut_RST = 0;
    #100000;
    $finish;
  end


endmodule



module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST
);

  reg [32-1:0] a;
  reg [32-1:0] b;
  reg [32-1:0] c;
  reg [32-1:0] d;
  reg signed [32-1:0] sa;
  reg signed [32-1:0] sb;
  reg signed [32-1:0] sc;
  reg signed [32-1:0] sd;

  always @(posedge CLK) begin
    if(RST) begin
      a <= 131072;
      b <= 4096;
      c <= 65536;
      d <= 65536;
      sa <= 131072;
      sb <= -4096;
      sc <= -65536;
      sd <= -65536;
    end else begin
      a <= a;
      b <= b;
      c <= (b << 8) / a << 16;
      d <= a / b << 8;
      sa <= sa;
      sb <= sb;
      sc <= (((((sb << 8) >> 31) & 1'b1) == sa[31])? ((!(((sb << 8) >> 31) & 1'b1))? sb << 8 : ~(sb << 8) + 1) / ((!sa[31])? sa : ~sa + 1) : ~(((!(((sb << 8) >> 31) & 1'b1))? sb << 8 : ~(sb << 8) + 1) / ((!sa[31])? sa : ~sa + 1)) + 1) << 16;
      sd <= ((sa[31] == sb[31])? ((!sa[31])? sa : ~sa + 1) / ((!sb[31])? sb : ~sb + 1) : ~(((!sa[31])? sa : ~sa + 1) / ((!sb[31])? sb : ~sb + 1)) + 1) << 8;
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_fixed_div.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import equals

expected_verilog = """
module blinkled
(
  input a,
  input b,
  input c
);

  assign c = a + b + 1 + (a + b + 1) + (a - b);

endmodule
"""


def test():
    veriloggen.reset()
    test_module = equals.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import submodule_read_verilog

expected_verilog = """
module top #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output [WIDTH-1:0] LED,
  output [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_out0,
  input [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_in0
);

  localparam inst_blinkled_WIDTH = WIDTH;
  wire [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_out2;
  reg [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_in2;
  wire [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_out1;
  wire [inst_blinkled_WIDTH-1+1-1:0] inst_blinkled_dummy_in1;

  blinkled
  #(
    .WIDTH(inst_blinkled_WIDTH)
  )
  inst_blinkled
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED),
    .dummy_out0(inst_blinkled_dummy_out0),
    .dummy_out1(inst_blinkled_dummy_out1),
    .dummy_out2(inst_blinkled_dummy_out2),
    .dummy_in0(inst_blinkled_dummy_in0),
    .dummy_in1(inst_blinkled_dummy_in1),
    .dummy_in2(inst_blinkled_dummy_in2)
  );


endmodule



module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED,
  output [WIDTH-1:0] dummy_out0,
  output [WIDTH-1:0] dummy_out1,
  output [WIDTH-1:0] dummy_out2,
  input [WIDTH-1:0] dummy_in0,
  input [WIDTH-1:0] dummy_in1,
  input [WIDTH-1:0] dummy_in2
);

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      if(count == 1023) begin
        LED <= LED + 1;
      end 
    end
  end


endmodule
"""


def test():
    test_module = submodule_read_verilog.mkTop()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from setuptools import setup, find_packages

import re
import os


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename), encoding='utf8').read()


m = re.search(r'(\d+\.\d+\.\d+(-.+)?)',
              read('veriloggen/utils/VERSION').splitlines()[0])
version = m.group(1) if m is not None else '0.0.0'

setup(name='veriloggen',
      version=version,
      description='A library for constructing a Verilog HDL source code in Python',
      long_description=read('README.rst'),
      keywords='FPGA, Verilog HDL, High-Level Synthesis',
      author='Shinya Takamaeda-Yamazaki',
      license="Apache License 2.0",
      url='https://github.com/PyHDI/veriloggen',
      packages=find_packages(),
      package_data={'veriloggen.utils': ['VERSION'],
                    'veriloggen.simulation': ['*.cpp'], },
      install_requires=['pyverilog>=1.1.4',
                        'Jinja2>=2.10',
                        'numpy>=1.14'],
      extras_require={
          'test': ['pytest>=3.2', 'pytest-pythonpath>=0.7'],
          'graph': ['pygraphviz>=1.3.1'],
      },
      )



from __future__ import absolute_import
from __future__ import print_function
import os
import sys

import veriloggen.core.vtypes as vtypes
from veriloggen.fsm.fsm import FSM

from .thread import Thread


def to_thread_pool(*threads):
    if not threads:
        raise ValueError('no thread specified.')

    if len(threads) == 1 and isinstance(threads[0], (tuple, list)):
        threads = threads[0]

    return ThreadPool(threads=threads)


class ThreadPool(vtypes.VeriloggenNode):
    __intrinsics__ = ('run', 'join', 'done', 'reset', 'ret')

    def __init__(self, m=None, name=None, clk=None, rst=None,
                 targ=None, numthreads=None, datawidth=32,
                 threads=None, fsm_as_module=False):

        if threads is not None:
            if not isinstance(threads, (tuple, list)):
                threads = [threads]

            clk = threads[0].clk
            rst = threads[0].rst
            name = threads[0].name
            self.m = threads[0].m
            self.numthreads = len(threads)
            self.threads = list(threads)
            for i, thread in enumerate(self.threads):
                thread.tid = i

        elif (m is not None and name is not None and
              clk is not None and rst is not None and
              targ is not None and numthreads is not None):

            self.m = m
            self.numthreads = numthreads
            self.threads = [Thread(m, '_'.join([name, str(i)]), clk, rst, targ,
                                   datawidth=datawidth, tid=i,
                                   fsm_as_module=fsm_as_module)
                            for i in range(numthreads)]

        else:
            raise ValueError('threads or other options must be specified.')

        self.start = self.m.Reg(
            '_'.join(['', name, 'start']), self.numthreads, initval=0)

    def start(self, tid, *args, **kwargs):
        return self.threads[tid].start(*args, **kwargs)

    def extend(self, tid, fsm, *args, **kwargs):
        return self.threads[tid].extend(fsm, *args, **kwargs)

    def run(self, fsm, tid, *args, **kwargs):
        """ start as a child thread """

        fsm(
            self.start[tid](1)
        )
        fsm.goto_next()

        for thread in self.threads:
            start_cond = self.start[thread.tid]

            if not thread.is_child and thread.end_state is not None:
                raise ValueError('thread %d already started' % thread.tid)

            if thread.fsm is None:
                thread.fsm = FSM(thread.m, thread.name, thread.clk, thread.rst,
                                 as_module=thread.fsm_as_module)

            thread.is_child = True

            if thread.start_state is not None:
                thread.fsm._set_index(thread.start_state)
            else:
                thread.start_state = thread.fsm.current

        fsm_start = fsm.current

        for thread in self.threads:
            start_cond = self.start[thread.tid]

            fsm._set_index(fsm_start)

            thread._synthesize_run_fsm(fsm, args, kwargs, cond=start_cond)

            if thread.end_state is None:
                thread.end_state = thread.fsm.current

        fsm(
            self.start[tid](0)
        )
        fsm.goto_next()

        return 0

    def join(self, fsm, tid):
        """ wait for the completion """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            end_flag = (thread.fsm.state == thread.end_state)
            patterns.append(((tid == thread.tid), end_flag))

        patterns.append((None, 0))
        cond = vtypes.PatternMux(*patterns)
        fsm.If(cond).goto_next()

        return 0

    def done(self, fsm, tid):
        """ check whethe the thread is running """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            end_flag = (thread.fsm.state == thread.end_state)
            patterns.append(((tid == thread.tid), end_flag))

        patterns.append((None, 0))
        cond = vtypes.PatternMux(*patterns)

        return cond

    def reset(self, fsm, tid):
        """ reset the FSM counter to the initial state """

        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)

        for thread in self.threads:
            reset_flag = vtypes.Land(
                (fsm.state == fsm.current), (tid == thread.tid))

            thread.fsm._set_index(thread.end_state)

            if thread.called is not None:
                thread.fsm.If(reset_flag)(
                    thread.called(0)
                )

            thread.fsm.goto_from(
                thread.end_state, thread.start_state, reset_flag)
            thread.fsm._set_index(thread.start_state)

        return 0

    def ret(self, fsm, tid):
        """ return value """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            patterns.append(((tid == thread.tid), thread.return_value))

        patterns.append((None, vtypes.IntX()))
        ret = vtypes.PatternMux(*patterns)

        return ret



from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import subprocess
import tempfile
import collections
from jinja2 import Environment, FileSystemLoader
try:
    from math import gcd
except:
    from fractions import gcd

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module
import veriloggen.core.submodule as submodule


def setup_waveform(m, *uuts, **kwargs):
    new_uuts = []
    for uut in uuts:
        if isinstance(uut, (tuple, list)):
            for u in uut:
                if isinstance(u, vtypes._Variable) and u.length is not None:
                    continue
                new_uuts.append(u)
        elif isinstance(uut, dict):
            _uut = list(uut.values())
            for u in _uut:
                if isinstance(u, vtypes._Variable) and u.length is not None:
                    continue
                new_uuts.append(u)
        elif isinstance(uut, submodule.Submodule):
            new_uuts.append(uut.inst)
        else:
            if isinstance(uut, vtypes._Variable) and uut.length is not None:
                continue
            new_uuts.append(uut)

    dumpfile = kwargs['dumpfile'] if 'dumpfile' in kwargs else 'uut.vcd'

    uuts = new_uuts
    ret = m.Initial(
        vtypes.Systask('dumpfile', dumpfile),
        vtypes.Systask('dumpvars', 0, *uuts)
    )

    # for verilator
    m.verilator_dumpfile = dumpfile

    return ret


def setup_clock(m, clk, hperiod=5):
    ret = m.Initial(
        clk(0),
        vtypes.Forever(clk(vtypes.Not(clk), ldelay=hperiod))
    )

    # for verilator
    if not hasattr(m, 'verilator_clock'):
        m.verilator_clock = collections.OrderedDict()

    m.verilator_clock[clk] = hperiod

    return ret


def setup_reset(m, reset, *statement, **kwargs):
    period = kwargs['period'] if 'period' in kwargs else 100
    polarity = kwargs['polarity'] if 'polarity' in kwargs else 'high'
    positive = 'high' in polarity.lower()

    ret = m.Initial(
        reset(not positive),
        statement,
        vtypes.Delay(period),
        reset(positive),
        vtypes.Delay(period),
        reset(not positive),
    )

    # for verilator
    if not hasattr(m, 'verilator_reset'):
        m.verilator_reset = collections.OrderedDict()

    m.verilator_reset[reset] = (period, positive)

    if not hasattr(m, 'verilator_reset_statements'):
        m.verilator_reset_statements = collections.OrderedDict()

    if reset not in m.verilator_reset_statements:
        m.verilator_reset_statements[reset] = []

    m.verilator_reset_statements[reset].extend(statement)

    return ret


def next_clock(clk):
    return (vtypes.Event(vtypes.Posedge(clk)), vtypes.Delay(1))


def finish():
    return vtypes.Systask('finish')


def _to_code(objs):
    code = []
    for obj in objs:
        if isinstance(obj, module.Module):
            code.append(obj.to_verilog())
            code.append('\n')
        if isinstance(obj, str):
            code.append(obj)
            code.append('\n')
    return ''.join(code)


def run_iverilog(objs, display=False, top=None, outputfile=None,
                 include=None, define=None, libdir=None):

    if not isinstance(objs, (tuple, list)):
        objs = [objs]

    if top is None:
        top = objs[0]

    top_name = top.name if isinstance(top, module.Module) else top

    if outputfile is None:
        outputfile = 'a.out'

    cmd = []
    cmd.append('iverilog')
    if include:
        for inc in include:
            cmd.append('-I')
            cmd.append(inc)

    if define:
        for d in define:
            cmd.append('-D')
            if isinstance(d, (tuple, list)):
                if d[1] is None:
                    cmd.append(d[0])
                else:
                    cmd.append(''.join([d[0], '=', str(d[1])]))
            else:
                cmd.append(d)

    if libdir:
        for l in libdir:
            cmd.append('-y')
            cmd.append(l)

    cmd.append('-s')
    cmd.append(top_name)

    cmd.append('-o')
    cmd.append(outputfile)

    # encoding: 'utf-8' ?
    encoding = sys.getdefaultencoding()

    code = _to_code(objs)

    if os.name == 'nt':
        fd, path = tempfile.mkstemp()
        os.close(fd)
        tmp = open(path, 'w+b')
    else:
        tmp = tempfile.NamedTemporaryFile()

    tmp.write(code.encode(encoding))
    tmp.read()
    filename = tmp.name

    cmd.append(filename)

    # synthesis
    syn_rslt = _exec(' '.join(cmd), encoding, display)

    # simulation
    cmd = []
    cmd.append('vvp')
    cmd.append(outputfile)
    sim_rslt = _exec(' '.join(cmd), encoding, display)

    # close temporal source code file
    tmp.close()

    if os.name == 'nt':
        os.remove(path)

    return ''.join([syn_rslt, sim_rslt])


def run_vcs(objs, display=False, top=None, outputfile=None,
            include=None, define=None, libdir=None,
            full64=False, notimingcheck=True):

    if not isinstance(objs, (tuple, list)):
        objs = [objs]

    if top is None:
        top = objs[0]

    top_name = top.name if isinstance(top, module.Module) else top

    if outputfile is None:
        outputfile = 'simv'

    cmd = []
    cmd.append('vcs')
    cmd.append('-v2005')

    if full64:
        cmd.append('-full64')

    if notimingcheck:
        cmd.append('+notimingcheck')

    if include:
        for inc in include:
            cmd.append('+incdir+')
            cmd.append(inc)

    if define:
        for d in define:
            cmd.append('+define+')
            if isinstance(d, (tuple, list)):
                if d[1] is None:
                    cmd.append(d[0])
                else:
                    cmd.append(''.join([d[0], '=', str(d[1])]))
            else:
                cmd.append(d)

    if libdir:
        for l in libdir:
            cmd.append('-y')
            cmd.append(l)

    cmd.append('-o')
    cmd.append(outputfile)

    # encoding: 'utf-8' ?
    encoding = sys.getdefaultencoding()

    code = _to_code(objs)

    if os.name == 'nt':
        fd, path = tempfile.mkstemp()
        os.close(fd)
        tmp = open(path, 'w+b')
    else:
        tmp = tempfile.NamedTemporaryFile()

    tmp.write(code.encode(encoding))
    tmp.read()
    filename = tmp.name

    cmd.append(filename)

    # synthesis
    syn_rslt = _exec(' '.join(cmd), encoding, display)

    # simulation
    cmd = []

    if os.name == 'nt':
        cmd.append(outputfile)
    else:
        cmd.append('./' + outputfile)

    sim_rslt = _exec(' '.join(cmd), encoding, display)

    # close temporal source code file
    tmp.close()

    if os.name == 'nt':
        os.remove(path)

    return ''.join([syn_rslt, sim_rslt])


def run_modelsim(objs, display=False, top=None, outputfile=None,
                 include=None, define=None, libdir=None):

    if not isinstance(objs, (tuple, list)):
        objs = [objs]

    if top is None:
        top = objs[0]

    top_name = top.name if isinstance(top, module.Module) else top

    if outputfile is None:
        outputfile = 'a.out'

    cmd = []
    cmd.append('vlib work ; vmap work ; vlog')

    if include:
        for inc in include:
            cmd.append('-I')
            cmd.append(inc)

    if define:
        for d in define:
            cmd.append('-D')
            if isinstance(d, (tuple, list)):
                if d[1] is None:
                    cmd.append(d[0])
                else:
                    cmd.append(''.join([d[0], '=', str(d[1])]))
            else:
                cmd.append(d)

    if libdir:
        for l in libdir:
            cmd.append('-y')
            cmd.append(l)

    # encoding: 'utf-8' ?
    encoding = sys.getdefaultencoding()

    code = _to_code(objs)

    if os.name == 'nt':
        fd, path = tempfile.mkstemp()
        os.close(fd)
        tmp = open(path, 'w+b')
    else:
        tmp = tempfile.NamedTemporaryFile()

    tmp.write(code.encode(encoding))
    tmp.read()
    filename = tmp.name

    cmd.append(filename)

    # synthesis
    syn_rslt = _exec(' '.join(cmd), encoding, display)

    # simulation
    cmd = []
    cmd.append('vsim -c')
    cmd.append(top)
    cmd.append('-do \"run -all\"')
    sim_rslt = _exec(' '.join(cmd), encoding, display)

    # close temporal source code file
    tmp.close()

    if os.name == 'nt':
        os.remove(path)

    return ''.join([syn_rslt, sim_rslt])


def to_verilator_code(top, objs):
    code = []

    if hasattr(top, 'verilator_clock'):
        clks = top.verilator_clock
    else:
        clks = {}

    if hasattr(top, 'verilator_reset'):
        rsts = top.verilator_reset
    else:
        rsts = {}

    if hasattr(top, 'verilator_reset_statements'):
        inits_list = top.verilator_reset_statements
        inits = collections.OrderedDict()
        for init in inits_list:
            if isinstance(init, vtypes.Subst):
                inits[init.left] = init.right
    else:
        inits = {}

    new_clks = collections.OrderedDict()
    for clk, values in clks.items():
        new_clk = top.InputLike(clk, name='io_' + clk.name)
        clk.connect(new_clk)
        new_clks[new_clk] = values

    top.verilator_new_clock = new_clks

    new_rsts = collections.OrderedDict()
    for rst, values in rsts.items():
        new_rst = top.InputLike(rst, name='io_' + rst.name)
        rst.connect(new_rst)
        new_rsts[new_rst] = values

    top.verilator_new_reset = new_rsts

    for obj in objs:
        if isinstance(obj, module.Module):
            code.append(obj.to_verilog(for_verilator=True))
            code.append('\n')
        if isinstance(obj, str):
            code.append(obj)
            code.append('\n')
    return ''.join(code)


def to_verilator_cpp(top, verilog_prefix, sim_time=0):
    template_path = os.path.dirname(os.path.abspath(__file__))
    env = Environment(loader=FileSystemLoader(template_path))
    env.globals['zip'] = zip

    if hasattr(top, 'verilator_dumpfile'):
        dumpfile = top.verilator_dumpfile
    else:
        dumpfile = None

    clks = top.verilator_new_clock
    rsts = top.verilator_new_reset

    if hasattr(top, 'verilator_reset_statements'):
        inits_list = top.verilator_reset_statements
        inits = collections.OrderedDict()
        for init in inits_list:
            if isinstance(init, vtypes.Subst):
                inits[init.left] = init.right
    else:
        inits = {}

    ios = top.get_ports()
    inputs = [io_var for io_var in ios.values()
              if isinstance(io_var, vtypes.Input) and
              (io_var not in clks) and (io_var not in rsts)]

    time_step = None

    for hperiod in clks.values():
        if time_step is None:
            time_step = hperiod
        else:
            time_step = gcd(time_step, hperiod)

    for period, positive in rsts.values():
        if time_step is None:
            time_step = period
        else:
            time_step = gcd(time_step, period)

    if time_step is None:
        time_step = 1

    template_dict = {
        'verilog_prefix': verilog_prefix,
        'sim_time': sim_time,
        'time_step': time_step,
        'dumpfile': dumpfile,
        'clks': clks,
        'rsts': rsts,
        'inits': inits,
        'inputs': inputs,
    }

    template = env.get_template('verilator_template.cpp')
    code = template.render(template_dict)
    return code


def run_verilator(objs, display=False, top=None, outputfile=None,
                  include=None, define=None, libdir=None,
                  sim_time=0, options=None):

    if not isinstance(objs, (tuple, list)):
        objs = [objs]

    if top is None:
        top = objs[0]

    top_name = top.name if isinstance(top, module.Module) else top

    if not isinstance(top, module.Module):
        for obj in objs:
            if obj.name == top_name:
                top = obj
                break

    if not isinstance(top, module.Module):
        raise ValueError('top module must be specified.')

    if outputfile is None:
        outputfile = 'obj_dir'

    cmd = []
    cmd.append('verilator')
    cmd.append('--cc')

    if options is not None:
        if isinstance(options, (tuple, list)):
            options = (options,)
        cmd.append(options)

    if hasattr(top, 'verilator_dumpfile'):
        cmd.append('--trace')
        cmd.append('--trace-underscore')

    # cmd.append('-Wall')
    cmd.append('-Wno-lint')

    if include:
        for inc in include:
            cmd.append('+incdir+')
            cmd.append(inc)

    if define:
        for d in define:
            cmd.append('+define+')
            if isinstance(d, (tuple, list)):
                if d[1] is None:
                    cmd.append(d[0])
                else:
                    cmd.append(''.join([d[0], '=', str(d[1])]))
            else:
                cmd.append(d)

    if libdir:
        for l in libdir:
            cmd.append('-y')
            cmd.append(l)

    cmd.append('--Mdir')
    cmd.append(outputfile)

    verilog_prefix = 'out'
    verilog_name = verilog_prefix + '.v'
    verilog_path = outputfile + '/' + verilog_name
    cmd.append(verilog_path)

    cpp_prefix = 'sim_%s' % top_name
    cpp_name = cpp_prefix + '.cpp'
    cpp_path = outputfile + '/' + cpp_name
    cmd.append('--exe')
    cmd.append(cpp_path)

    to_verilator(top, objs,
                 sim_time, outputfile, verilog_prefix, cpp_prefix)

    # for clk in top.verilator_clock.keys():
    #    cmd.append('--clk')
    #    cmd.append(clk.name)

    # for clk in top.verilator_new_clock.keys():
    #    cmd.append('--clk')
    #    cmd.append(clk.name)

    # encoding: 'utf-8' ?
    encoding = sys.getdefaultencoding()

    # synthesis
    syn_rslt = _exec(' '.join(cmd), encoding, display)

    # make
    make = ['make -C', outputfile, '-j -f',
            'V' + verilog_prefix + '.mk', 'V' + verilog_prefix]
    make_rslt = _exec(' '.join(make), encoding, display)

    # simulation
    cmd = []

    if os.name == 'nt':
        cmd.append(outputfile + '/' + 'V' + verilog_prefix)
    else:
        cmd.append('./' + outputfile + '/' + 'V' + verilog_prefix)

    sim_rslt = _exec(' '.join(cmd), encoding, display)

    # return ''.join([syn_rslt, make_rslt, sim_rslt])
    return sim_rslt


def to_verilator(top, objs,
                 sim_time=0, outputdir='obj_dir',
                 verilog_prefix='out', cpp_prefix='sim'):

    if not os.path.exists(outputdir):
        os.mkdir(outputdir)

    verilog_name = verilog_prefix + '.v'
    verilog_path = outputdir + '/' + verilog_name

    verilog_code = to_verilator_code(top, objs)
    verilog = open(verilog_path, 'w')
    verilog.write(verilog_code)
    verilog.close()

    cpp_name = cpp_prefix + '.cpp'
    cpp_path = outputdir + '/' + cpp_name

    cpp_code = to_verilator_cpp(top, verilog_prefix, sim_time)
    cpp = open(cpp_path, 'w')
    cpp.write(cpp_code)
    cpp.close()


def view_waveform_gtkwave(filename='uut.vcd', background=False):
    cmd = []
    cmd.append('gtkwave')
    cmd.append('--giga')
    cmd.append(filename)
    if background:
        cmd.append('&')
    subprocess.call(' '.join(cmd), shell=True)


class Simulator(object):
    """ obsoleted class """

    def __init__(self, *objs, **options):
        if not objs:
            raise ValueError('no target object')

        sim = 'iverilog' if 'sim' not in options else options['sim'].lower()
        wave = 'gtkwave' if 'wave' not in options else options['wave'].lower()

        # for VCS
        full64 = False if 'full64' not in options else options['full64']
        notimingcheck = True if 'notimingcheck' not in options else options['notimingcheck']

        self.objs = objs
        self.sim = sim
        self.wave = wave

        self.full64 = full64
        self.notimingcheck = notimingcheck

    def run(self, display=False, top=None, outputfile=None,
            include=None, define=None, libdir=None,
            sim_time=0, options=None):

        if include is not None and not isinstance(include, (tuple, list)):
            include = (include,)

        if define is not None and not isinstance(define, (tuple, list)):
            define = (define,)

        if libdir is not None and not isinstance(libdir, (tuple, list)):
            libdir = (libdir,)

        if self.sim == 'iverilog' or self.sim == 'icarus':
            return run_iverilog(self.objs,
                                display, top, outputfile, include, define, libdir)

        if self.sim == 'vcs':
            return run_vcs(self.objs,
                           display, top, outputfile, include, define, libdir,
                           full64=self.full64, notimingcheck=self.notimingcheck)

        if self.sim == 'modelsim' or self.sim == 'vsim':
            return run_modelsim(self.objs,
                                display, top, include, define, libdir)

        if self.sim == 'verilator':
            return run_verilator(self.objs,
                                 display, top, outputfile, include, define, libdir,
                                 sim_time=sim_time, options=options)

        raise NotImplementedError("not supported simulator: '%s'" % self.sim)

    def view_waveform(self, filename='uut.vcd', background=False):
        if self.wave == 'gtkwave':
            return view_waveform_gtkwave(filename, background)

        raise NotImplementedError(
            "not supported waveform viewer: '%s'" % self.wave)


def _exec(cmd, encoding, display=False):
    if display:
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        rslt = []

        while True:
            stdout_data = p.stdout.readline()
            out = stdout_data.decode(encoding)
            rslt.append(out)
            print(out, end='')
            if not stdout_data:
                break

        p.wait()
        p.stdout.close()
        rslt = ''.join(rslt)

    else:
        b = subprocess.check_output(cmd, shell=True)
        rslt = b.decode(encoding)

    return rslt



from __future__ import absolute_import
from __future__ import print_function

import os
import veriloggen
import thread_stream_fixed


def test(request):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = thread_stream_fixed.run(filename=None, simtype=simtype,
                                   outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = rslt.splitlines()[-1]
    assert(verify_rslt == '# verify: PASSED')



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_axi_slave_write_lite

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [32-1:0] sum;
  reg [32-1:0] myaxi_awaddr;
  reg [4-1:0] myaxi_awcache;
  reg [3-1:0] myaxi_awprot;
  reg myaxi_awvalid;
  wire myaxi_awready;
  reg [32-1:0] myaxi_wdata;
  reg [4-1:0] myaxi_wstrb;
  reg myaxi_wvalid;
  wire myaxi_wready;
  wire [2-1:0] myaxi_bresp;
  wire myaxi_bvalid;
  reg myaxi_bready;
  reg [32-1:0] myaxi_araddr;
  reg [4-1:0] myaxi_arcache;
  reg [3-1:0] myaxi_arprot;
  reg myaxi_arvalid;
  wire myaxi_arready;
  wire [32-1:0] myaxi_rdata;
  wire [2-1:0] myaxi_rresp;
  wire myaxi_rvalid;
  reg myaxi_rready;
  reg [32-1:0] _axi_awaddr;
  wire [4-1:0] _axi_awcache;
  wire [3-1:0] _axi_awprot;
  reg _axi_awvalid;
  wire _axi_awready;
  reg [32-1:0] _axi_wdata;
  reg [4-1:0] _axi_wstrb;
  reg _axi_wvalid;
  wire _axi_wready;
  wire [2-1:0] _axi_bresp;
  wire _axi_bvalid;
  wire _axi_bready;
  reg [32-1:0] _axi_araddr;
  wire [4-1:0] _axi_arcache;
  wire [3-1:0] _axi_arprot;
  reg _axi_arvalid;
  wire _axi_arready;
  wire [32-1:0] _axi_rdata;
  wire [2-1:0] _axi_rresp;
  wire _axi_rvalid;
  wire _axi_rready;
  assign _axi_awcache = 3;
  assign _axi_awprot = 0;
  assign _axi_bready = 1;
  assign _axi_arcache = 3;
  assign _axi_arprot = 0;
  assign _axi_rready = 0;
  wire [32-1:0] _tmp_0;
  assign _tmp_0 = _axi_awaddr;

  always @(*) begin
    myaxi_awaddr = _tmp_0;
  end

  wire [4-1:0] _tmp_1;
  assign _tmp_1 = _axi_awcache;

  always @(*) begin
    myaxi_awcache = _tmp_1;
  end

  wire [3-1:0] _tmp_2;
  assign _tmp_2 = _axi_awprot;

  always @(*) begin
    myaxi_awprot = _tmp_2;
  end

  wire _tmp_3;
  assign _tmp_3 = _axi_awvalid;

  always @(*) begin
    myaxi_awvalid = _tmp_3;
  end

  assign _axi_awready = myaxi_awready;
  wire [32-1:0] _tmp_4;
  assign _tmp_4 = _axi_wdata;

  always @(*) begin
    myaxi_wdata = _tmp_4;
  end

  wire [4-1:0] _tmp_5;
  assign _tmp_5 = _axi_wstrb;

  always @(*) begin
    myaxi_wstrb = _tmp_5;
  end

  wire _tmp_6;
  assign _tmp_6 = _axi_wvalid;

  always @(*) begin
    myaxi_wvalid = _tmp_6;
  end

  assign _axi_wready = myaxi_wready;
  assign _axi_bresp = myaxi_bresp;
  assign _axi_bvalid = myaxi_bvalid;
  wire _tmp_7;
  assign _tmp_7 = _axi_bready;

  always @(*) begin
    myaxi_bready = _tmp_7;
  end

  wire [32-1:0] _tmp_8;
  assign _tmp_8 = _axi_araddr;

  always @(*) begin
    myaxi_araddr = _tmp_8;
  end

  wire [4-1:0] _tmp_9;
  assign _tmp_9 = _axi_arcache;

  always @(*) begin
    myaxi_arcache = _tmp_9;
  end

  wire [3-1:0] _tmp_10;
  assign _tmp_10 = _axi_arprot;

  always @(*) begin
    myaxi_arprot = _tmp_10;
  end

  wire _tmp_11;
  assign _tmp_11 = _axi_arvalid;

  always @(*) begin
    myaxi_arvalid = _tmp_11;
  end

  assign _axi_arready = myaxi_arready;
  assign _axi_rdata = myaxi_rdata;
  assign _axi_rresp = myaxi_rresp;
  assign _axi_rvalid = myaxi_rvalid;
  wire _tmp_12;
  assign _tmp_12 = _axi_rready;

  always @(*) begin
    myaxi_rready = _tmp_12;
  end

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg __axi_cond_0_1;
  reg [32-1:0] wdata;
  reg __axi_cond_1_1;
  reg __axi_cond_2_1;
  reg __axi_cond_3_1;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .sum(sum),
    .myaxi_awaddr(myaxi_awaddr),
    .myaxi_awcache(myaxi_awcache),
    .myaxi_awprot(myaxi_awprot),
    .myaxi_awvalid(myaxi_awvalid),
    .myaxi_awready(myaxi_awready),
    .myaxi_wdata(myaxi_wdata),
    .myaxi_wstrb(myaxi_wstrb),
    .myaxi_wvalid(myaxi_wvalid),
    .myaxi_wready(myaxi_wready),
    .myaxi_bresp(myaxi_bresp),
    .myaxi_bvalid(myaxi_bvalid),
    .myaxi_bready(myaxi_bready),
    .myaxi_araddr(myaxi_araddr),
    .myaxi_arcache(myaxi_arcache),
    .myaxi_arprot(myaxi_arprot),
    .myaxi_arvalid(myaxi_arvalid),
    .myaxi_arready(myaxi_arready),
    .myaxi_rdata(myaxi_rdata),
    .myaxi_rresp(myaxi_rresp),
    .myaxi_rvalid(myaxi_rvalid),
    .myaxi_rready(myaxi_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, sum, myaxi_awaddr, myaxi_awcache, myaxi_awprot, myaxi_awvalid, myaxi_awready, myaxi_wdata, myaxi_wstrb, myaxi_wvalid, myaxi_wready, myaxi_bresp, myaxi_bvalid, myaxi_bready, myaxi_araddr, myaxi_arcache, myaxi_arprot, myaxi_arvalid, myaxi_arready, myaxi_rdata, myaxi_rresp, myaxi_rvalid, myaxi_rready, _axi_awaddr, _axi_awcache, _axi_awprot, _axi_awvalid, _axi_awready, _axi_wdata, _axi_wstrb, _axi_wvalid, _axi_wready, _axi_bresp, _axi_bvalid, _axi_bready, _axi_araddr, _axi_arcache, _axi_arprot, _axi_arvalid, _axi_arready, _axi_rdata, _axi_rresp, _axi_rvalid, _axi_rready, _tmp_0, _tmp_1, _tmp_2, _tmp_3, _tmp_4, _tmp_5, _tmp_6, _tmp_7, _tmp_8, _tmp_9, _tmp_10, _tmp_11, _tmp_12, fsm, __axi_cond_0_1, wdata, __axi_cond_1_1, __axi_cond_2_1, __axi_cond_3_1);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    _axi_awaddr = 0;
    _axi_awvalid = 0;
    _axi_wdata = 0;
    _axi_wstrb = 0;
    _axi_wvalid = 0;
    _axi_araddr = 0;
    _axi_arvalid = 0;
    fsm = fsm_init;
    __axi_cond_0_1 = 0;
    wdata = 0;
    __axi_cond_1_1 = 0;
    __axi_cond_2_1 = 0;
    __axi_cond_3_1 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


  always @(posedge CLK) begin
    if(RST) begin
      _axi_araddr <= 0;
      _axi_arvalid <= 0;
      _axi_awaddr <= 0;
      _axi_awvalid <= 0;
      __axi_cond_0_1 <= 0;
      _axi_wdata <= 0;
      _axi_wvalid <= 0;
      _axi_wstrb <= 0;
      __axi_cond_1_1 <= 0;
      __axi_cond_2_1 <= 0;
      __axi_cond_3_1 <= 0;
    end else begin
      if(__axi_cond_0_1) begin
        _axi_awvalid <= 0;
      end 
      if(__axi_cond_1_1) begin
        _axi_wvalid <= 0;
      end 
      if(__axi_cond_2_1) begin
        _axi_awvalid <= 0;
      end 
      if(__axi_cond_3_1) begin
        _axi_wvalid <= 0;
      end 
      _axi_araddr <= 0;
      _axi_arvalid <= 0;
      if((fsm == 100) && (_axi_awready || !_axi_awvalid)) begin
        _axi_awaddr <= 1024;
        _axi_awvalid <= 1;
      end 
      __axi_cond_0_1 <= 1;
      if(_axi_awvalid && !_axi_awready) begin
        _axi_awvalid <= _axi_awvalid;
      end 
      if((fsm == 101) && (_axi_wready || !_axi_wvalid)) begin
        _axi_wdata <= wdata;
        _axi_wvalid <= 1;
        _axi_wstrb <= { 4{ 1'd1 } };
      end 
      __axi_cond_1_1 <= 1;
      if(_axi_wvalid && !_axi_wready) begin
        _axi_wvalid <= _axi_wvalid;
      end 
      if((fsm == 102) && (_axi_awready || !_axi_awvalid)) begin
        _axi_awaddr <= 1024;
        _axi_awvalid <= 1;
      end 
      __axi_cond_2_1 <= 1;
      if(_axi_awvalid && !_axi_awready) begin
        _axi_awvalid <= _axi_awvalid;
      end 
      if((fsm == 103) && (_axi_wready || !_axi_wvalid)) begin
        _axi_wdata <= wdata;
        _axi_wvalid <= 1;
        _axi_wstrb <= { 4{ 1'd1 } };
      end 
      __axi_cond_3_1 <= 1;
      if(_axi_wvalid && !_axi_wready) begin
        _axi_wvalid <= _axi_wvalid;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;
  localparam fsm_8 = 8;
  localparam fsm_9 = 9;
  localparam fsm_10 = 10;
  localparam fsm_11 = 11;
  localparam fsm_12 = 12;
  localparam fsm_13 = 13;
  localparam fsm_14 = 14;
  localparam fsm_15 = 15;
  localparam fsm_16 = 16;
  localparam fsm_17 = 17;
  localparam fsm_18 = 18;
  localparam fsm_19 = 19;
  localparam fsm_20 = 20;
  localparam fsm_21 = 21;
  localparam fsm_22 = 22;
  localparam fsm_23 = 23;
  localparam fsm_24 = 24;
  localparam fsm_25 = 25;
  localparam fsm_26 = 26;
  localparam fsm_27 = 27;
  localparam fsm_28 = 28;
  localparam fsm_29 = 29;
  localparam fsm_30 = 30;
  localparam fsm_31 = 31;
  localparam fsm_32 = 32;
  localparam fsm_33 = 33;
  localparam fsm_34 = 34;
  localparam fsm_35 = 35;
  localparam fsm_36 = 36;
  localparam fsm_37 = 37;
  localparam fsm_38 = 38;
  localparam fsm_39 = 39;
  localparam fsm_40 = 40;
  localparam fsm_41 = 41;
  localparam fsm_42 = 42;
  localparam fsm_43 = 43;
  localparam fsm_44 = 44;
  localparam fsm_45 = 45;
  localparam fsm_46 = 46;
  localparam fsm_47 = 47;
  localparam fsm_48 = 48;
  localparam fsm_49 = 49;
  localparam fsm_50 = 50;
  localparam fsm_51 = 51;
  localparam fsm_52 = 52;
  localparam fsm_53 = 53;
  localparam fsm_54 = 54;
  localparam fsm_55 = 55;
  localparam fsm_56 = 56;
  localparam fsm_57 = 57;
  localparam fsm_58 = 58;
  localparam fsm_59 = 59;
  localparam fsm_60 = 60;
  localparam fsm_61 = 61;
  localparam fsm_62 = 62;
  localparam fsm_63 = 63;
  localparam fsm_64 = 64;
  localparam fsm_65 = 65;
  localparam fsm_66 = 66;
  localparam fsm_67 = 67;
  localparam fsm_68 = 68;
  localparam fsm_69 = 69;
  localparam fsm_70 = 70;
  localparam fsm_71 = 71;
  localparam fsm_72 = 72;
  localparam fsm_73 = 73;
  localparam fsm_74 = 74;
  localparam fsm_75 = 75;
  localparam fsm_76 = 76;
  localparam fsm_77 = 77;
  localparam fsm_78 = 78;
  localparam fsm_79 = 79;
  localparam fsm_80 = 80;
  localparam fsm_81 = 81;
  localparam fsm_82 = 82;
  localparam fsm_83 = 83;
  localparam fsm_84 = 84;
  localparam fsm_85 = 85;
  localparam fsm_86 = 86;
  localparam fsm_87 = 87;
  localparam fsm_88 = 88;
  localparam fsm_89 = 89;
  localparam fsm_90 = 90;
  localparam fsm_91 = 91;
  localparam fsm_92 = 92;
  localparam fsm_93 = 93;
  localparam fsm_94 = 94;
  localparam fsm_95 = 95;
  localparam fsm_96 = 96;
  localparam fsm_97 = 97;
  localparam fsm_98 = 98;
  localparam fsm_99 = 99;
  localparam fsm_100 = 100;
  localparam fsm_101 = 101;
  localparam fsm_102 = 102;
  localparam fsm_103 = 103;
  localparam fsm_104 = 104;
  localparam fsm_105 = 105;
  localparam fsm_106 = 106;
  localparam fsm_107 = 107;
  localparam fsm_108 = 108;
  localparam fsm_109 = 109;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      wdata <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          fsm <= fsm_1;
        end
        fsm_1: begin
          fsm <= fsm_2;
        end
        fsm_2: begin
          fsm <= fsm_3;
        end
        fsm_3: begin
          fsm <= fsm_4;
        end
        fsm_4: begin
          fsm <= fsm_5;
        end
        fsm_5: begin
          fsm <= fsm_6;
        end
        fsm_6: begin
          fsm <= fsm_7;
        end
        fsm_7: begin
          fsm <= fsm_8;
        end
        fsm_8: begin
          fsm <= fsm_9;
        end
        fsm_9: begin
          fsm <= fsm_10;
        end
        fsm_10: begin
          fsm <= fsm_11;
        end
        fsm_11: begin
          fsm <= fsm_12;
        end
        fsm_12: begin
          fsm <= fsm_13;
        end
        fsm_13: begin
          fsm <= fsm_14;
        end
        fsm_14: begin
          fsm <= fsm_15;
        end
        fsm_15: begin
          fsm <= fsm_16;
        end
        fsm_16: begin
          fsm <= fsm_17;
        end
        fsm_17: begin
          fsm <= fsm_18;
        end
        fsm_18: begin
          fsm <= fsm_19;
        end
        fsm_19: begin
          fsm <= fsm_20;
        end
        fsm_20: begin
          fsm <= fsm_21;
        end
        fsm_21: begin
          fsm <= fsm_22;
        end
        fsm_22: begin
          fsm <= fsm_23;
        end
        fsm_23: begin
          fsm <= fsm_24;
        end
        fsm_24: begin
          fsm <= fsm_25;
        end
        fsm_25: begin
          fsm <= fsm_26;
        end
        fsm_26: begin
          fsm <= fsm_27;
        end
        fsm_27: begin
          fsm <= fsm_28;
        end
        fsm_28: begin
          fsm <= fsm_29;
        end
        fsm_29: begin
          fsm <= fsm_30;
        end
        fsm_30: begin
          fsm <= fsm_31;
        end
        fsm_31: begin
          fsm <= fsm_32;
        end
        fsm_32: begin
          fsm <= fsm_33;
        end
        fsm_33: begin
          fsm <= fsm_34;
        end
        fsm_34: begin
          fsm <= fsm_35;
        end
        fsm_35: begin
          fsm <= fsm_36;
        end
        fsm_36: begin
          fsm <= fsm_37;
        end
        fsm_37: begin
          fsm <= fsm_38;
        end
        fsm_38: begin
          fsm <= fsm_39;
        end
        fsm_39: begin
          fsm <= fsm_40;
        end
        fsm_40: begin
          fsm <= fsm_41;
        end
        fsm_41: begin
          fsm <= fsm_42;
        end
        fsm_42: begin
          fsm <= fsm_43;
        end
        fsm_43: begin
          fsm <= fsm_44;
        end
        fsm_44: begin
          fsm <= fsm_45;
        end
        fsm_45: begin
          fsm <= fsm_46;
        end
        fsm_46: begin
          fsm <= fsm_47;
        end
        fsm_47: begin
          fsm <= fsm_48;
        end
        fsm_48: begin
          fsm <= fsm_49;
        end
        fsm_49: begin
          fsm <= fsm_50;
        end
        fsm_50: begin
          fsm <= fsm_51;
        end
        fsm_51: begin
          fsm <= fsm_52;
        end
        fsm_52: begin
          fsm <= fsm_53;
        end
        fsm_53: begin
          fsm <= fsm_54;
        end
        fsm_54: begin
          fsm <= fsm_55;
        end
        fsm_55: begin
          fsm <= fsm_56;
        end
        fsm_56: begin
          fsm <= fsm_57;
        end
        fsm_57: begin
          fsm <= fsm_58;
        end
        fsm_58: begin
          fsm <= fsm_59;
        end
        fsm_59: begin
          fsm <= fsm_60;
        end
        fsm_60: begin
          fsm <= fsm_61;
        end
        fsm_61: begin
          fsm <= fsm_62;
        end
        fsm_62: begin
          fsm <= fsm_63;
        end
        fsm_63: begin
          fsm <= fsm_64;
        end
        fsm_64: begin
          fsm <= fsm_65;
        end
        fsm_65: begin
          fsm <= fsm_66;
        end
        fsm_66: begin
          fsm <= fsm_67;
        end
        fsm_67: begin
          fsm <= fsm_68;
        end
        fsm_68: begin
          fsm <= fsm_69;
        end
        fsm_69: begin
          fsm <= fsm_70;
        end
        fsm_70: begin
          fsm <= fsm_71;
        end
        fsm_71: begin
          fsm <= fsm_72;
        end
        fsm_72: begin
          fsm <= fsm_73;
        end
        fsm_73: begin
          fsm <= fsm_74;
        end
        fsm_74: begin
          fsm <= fsm_75;
        end
        fsm_75: begin
          fsm <= fsm_76;
        end
        fsm_76: begin
          fsm <= fsm_77;
        end
        fsm_77: begin
          fsm <= fsm_78;
        end
        fsm_78: begin
          fsm <= fsm_79;
        end
        fsm_79: begin
          fsm <= fsm_80;
        end
        fsm_80: begin
          fsm <= fsm_81;
        end
        fsm_81: begin
          fsm <= fsm_82;
        end
        fsm_82: begin
          fsm <= fsm_83;
        end
        fsm_83: begin
          fsm <= fsm_84;
        end
        fsm_84: begin
          fsm <= fsm_85;
        end
        fsm_85: begin
          fsm <= fsm_86;
        end
        fsm_86: begin
          fsm <= fsm_87;
        end
        fsm_87: begin
          fsm <= fsm_88;
        end
        fsm_88: begin
          fsm <= fsm_89;
        end
        fsm_89: begin
          fsm <= fsm_90;
        end
        fsm_90: begin
          fsm <= fsm_91;
        end
        fsm_91: begin
          fsm <= fsm_92;
        end
        fsm_92: begin
          fsm <= fsm_93;
        end
        fsm_93: begin
          fsm <= fsm_94;
        end
        fsm_94: begin
          fsm <= fsm_95;
        end
        fsm_95: begin
          fsm <= fsm_96;
        end
        fsm_96: begin
          fsm <= fsm_97;
        end
        fsm_97: begin
          fsm <= fsm_98;
        end
        fsm_98: begin
          fsm <= fsm_99;
        end
        fsm_99: begin
          fsm <= fsm_100;
        end
        fsm_100: begin
          if(_axi_awready || !_axi_awvalid) begin
            wdata <= 100;
          end 
          if(_axi_awready || !_axi_awvalid) begin
            fsm <= fsm_101;
          end 
        end
        fsm_101: begin
          if(_axi_wready || !_axi_wvalid) begin
            fsm <= fsm_102;
          end 
        end
        fsm_102: begin
          if(_axi_awready || !_axi_awvalid) begin
            wdata <= 200;
          end 
          if(_axi_awready || !_axi_awvalid) begin
            fsm <= fsm_103;
          end 
        end
        fsm_103: begin
          if(_axi_wready || !_axi_wvalid) begin
            fsm <= fsm_104;
          end 
        end
        fsm_104: begin
          fsm <= fsm_105;
        end
        fsm_105: begin
          fsm <= fsm_106;
        end
        fsm_106: begin
          fsm <= fsm_107;
        end
        fsm_107: begin
          fsm <= fsm_108;
        end
        fsm_108: begin
          $display("sum=%d expected_sum=%d", sum, 300);
          fsm <= fsm_109;
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  output reg [32-1:0] sum,
  input [32-1:0] myaxi_awaddr,
  input [4-1:0] myaxi_awcache,
  input [3-1:0] myaxi_awprot,
  input myaxi_awvalid,
  output myaxi_awready,
  input [32-1:0] myaxi_wdata,
  input [4-1:0] myaxi_wstrb,
  input myaxi_wvalid,
  output myaxi_wready,
  output [2-1:0] myaxi_bresp,
  output reg myaxi_bvalid,
  input myaxi_bready,
  input [32-1:0] myaxi_araddr,
  input [4-1:0] myaxi_arcache,
  input [3-1:0] myaxi_arprot,
  input myaxi_arvalid,
  output myaxi_arready,
  output reg [32-1:0] myaxi_rdata,
  output [2-1:0] myaxi_rresp,
  output reg myaxi_rvalid,
  input myaxi_rready
);

  assign myaxi_bresp = 0;
  assign myaxi_rresp = 0;
  assign myaxi_arready = 0;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _tmp_0;
  reg _tmp_1;
  reg _tmp_2;
  assign myaxi_awready = (fsm == 0) && !_tmp_1 && !myaxi_bvalid && _tmp_2;
  assign myaxi_wready = fsm == 1;

  always @(posedge CLK) begin
    if(RST) begin
      myaxi_bvalid <= 0;
      myaxi_rvalid <= 0;
      _tmp_2 <= 0;
      _tmp_0 <= 0;
      _tmp_1 <= 0;
    end else begin
      if(myaxi_bvalid && myaxi_bready) begin
        myaxi_bvalid <= 0;
      end 
      if(myaxi_wvalid && myaxi_wready) begin
        myaxi_bvalid <= 1;
      end 
      myaxi_rvalid <= 0;
      _tmp_2 <= myaxi_awvalid;
      if(myaxi_awready && myaxi_awvalid && !myaxi_bvalid) begin
        _tmp_0 <= myaxi_awaddr;
      end 
      _tmp_1 <= myaxi_awready && myaxi_awvalid && !myaxi_bvalid;
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      sum <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(_tmp_1) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(myaxi_wready && myaxi_wvalid) begin
            sum <= sum + myaxi_wdata;
          end 
          if(myaxi_wready && myaxi_wvalid) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          fsm <= fsm_init;
        end
      endcase
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_axi_slave_write_lite.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import thread_ram_read_dataflow_reuse

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [14-1:0] myram_0_addr;
  wire [32-1:0] myram_0_rdata;
  reg [32-1:0] myram_0_wdata;
  reg myram_0_wenable;
  reg [14-1:0] myram_1_addr;
  wire [32-1:0] myram_1_rdata;
  reg [32-1:0] myram_1_wdata;
  reg myram_1_wenable;

  myram
  inst_myram
  (
    .CLK(CLK),
    .myram_0_addr(myram_0_addr),
    .myram_0_rdata(myram_0_rdata),
    .myram_0_wdata(myram_0_wdata),
    .myram_0_wenable(myram_0_wenable),
    .myram_1_addr(myram_1_addr),
    .myram_1_rdata(myram_1_rdata),
    .myram_1_wdata(myram_1_wdata),
    .myram_1_wenable(myram_1_wenable)
  );

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [8-1:0] _tmp_0;
  reg _tmp_1;
  wire [32-1:0] _dataflow_counter_odata_1;
  wire _dataflow_counter_ovalid_1;
  wire _dataflow_counter_oready_1;
  assign _dataflow_counter_oready_1 = (_tmp_0 > 0) && !_tmp_1;
  reg _myram_cond_0_1;
  reg _tmp_2;
  reg _tmp_3;
  reg _tmp_4;
  wire _tmp_5;
  wire _tmp_6;
  wire _tmp_7;
  reg [7-1:0] _tmp_8;
  reg _tmp_9;
  reg signed [32-1:0] _tmp_10;
  reg signed [32-1:0] _tmp_11;
  reg signed [32-1:0] _tmp_12;
  reg signed [32-1:0] _tmp_13;
  reg [4-1:0] _tmp_14;
  reg _tmp_15;
  reg _tmp_16;
  reg [32-1:0] _tmp_fsm_0;
  localparam _tmp_fsm_0_init = 0;
  reg [32-1:0] _d1__tmp_fsm_0;
  reg [32-1:0] _d2__tmp_fsm_0;
  reg __tmp_fsm_0_cond_1_0_1;
  reg __tmp_fsm_0_cond_1_0_2;
  reg __tmp_fsm_0_cond_2_1_1;
  reg __tmp_fsm_0_cond_2_1_2;
  reg __tmp_fsm_0_cond_5_2_1;
  reg __tmp_fsm_0_cond_6_3_1;
  reg __tmp_fsm_0_cond_6_3_2;
  reg __tmp_fsm_0_cond_7_4_1;
  reg __tmp_fsm_0_cond_7_4_2;
  reg __tmp_fsm_0_cond_10_5_1;
  wire signed [32-1:0] _dataflow__variable_odata_4;
  wire _dataflow__variable_ovalid_4;
  wire _dataflow__variable_oready_4;
  assign _dataflow__variable_oready_4 = 1;
  wire signed [32-1:0] _dataflow__variable_odata_5;
  wire _dataflow__variable_ovalid_5;
  wire _dataflow__variable_oready_5;
  assign _dataflow__variable_oready_5 = 1;
  wire [1-1:0] _dataflow__variable_odata_3;
  wire _dataflow__variable_ovalid_3;
  wire _dataflow__variable_oready_3;
  assign _dataflow__variable_oready_3 = 1;
  reg [32-1:0] sum0;
  reg [32-1:0] sum1;
  reg _seq_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      myram_1_wdata <= 0;
      myram_1_wenable <= 0;
      myram_0_addr <= 0;
      _tmp_0 <= 0;
      myram_0_wdata <= 0;
      myram_0_wenable <= 0;
      _tmp_1 <= 0;
      _myram_cond_0_1 <= 0;
      _tmp_9 <= 0;
      _tmp_2 <= 0;
      _tmp_3 <= 0;
      _tmp_4 <= 0;
      _tmp_14 <= 0;
    end else begin
      if(_myram_cond_0_1) begin
        myram_0_wenable <= 0;
        _tmp_1 <= 0;
      end 
      myram_1_wdata <= 0;
      myram_1_wenable <= 0;
      if((fsm == 0) && (_tmp_0 == 0)) begin
        myram_0_addr <= -1;
        _tmp_0 <= 64;
      end 
      if(_dataflow_counter_ovalid_1 && ((_tmp_0 > 0) && !_tmp_1) && (_tmp_0 > 0)) begin
        myram_0_addr <= myram_0_addr + 1;
        myram_0_wdata <= _dataflow_counter_odata_1;
        myram_0_wenable <= 1;
        _tmp_0 <= _tmp_0 - 1;
      end 
      if(_dataflow_counter_ovalid_1 && ((_tmp_0 > 0) && !_tmp_1) && (_tmp_0 == 1)) begin
        _tmp_1 <= 1;
      end 
      _myram_cond_0_1 <= 1;
      if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && _tmp_4) begin
        _tmp_9 <= 0;
        _tmp_2 <= 0;
        _tmp_3 <= 0;
        _tmp_4 <= 0;
      end 
      if(_tmp_15) begin
        _tmp_14 <= 4;
      end 
      if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 > 0)) begin
        _tmp_14 <= _tmp_14 - 1;
        _tmp_2 <= 1;
        _tmp_3 <= 1;
        _tmp_4 <= 1;
        _tmp_9 <= 0;
      end 
      if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 1) && _tmp_16) begin
        _tmp_9 <= 1;
      end 
    end
  end

  assign _dataflow__variable_odata_3 = _tmp_9;
  assign _dataflow__variable_ovalid_3 = _tmp_4;
  assign _tmp_7 = 1 && _dataflow__variable_oready_3;
  assign _dataflow__variable_odata_4 = _tmp_10;
  assign _dataflow__variable_ovalid_4 = _tmp_2;
  assign _tmp_5 = 1 && _dataflow__variable_oready_4;
  assign _dataflow__variable_odata_5 = _tmp_11;
  assign _dataflow__variable_ovalid_5 = _tmp_3;
  assign _tmp_6 = 1 && _dataflow__variable_oready_5;
  reg [32-1:0] _dataflow_counter_data_1;
  reg _dataflow_counter_valid_1;
  wire _dataflow_counter_ready_1;
  assign _dataflow_counter_odata_1 = _dataflow_counter_data_1;
  assign _dataflow_counter_ovalid_1 = _dataflow_counter_valid_1;
  assign _dataflow_counter_ready_1 = _dataflow_counter_oready_1;

  always @(posedge CLK) begin
    if(RST) begin
      _dataflow_counter_data_1 <= -2'sd1;
      _dataflow_counter_valid_1 <= 0;
    end else begin
      if((_dataflow_counter_ready_1 || !_dataflow_counter_valid_1) && 1 && 1) begin
        _dataflow_counter_data_1 <= _dataflow_counter_data_1 + 1;
      end 
      if(_dataflow_counter_valid_1 && _dataflow_counter_ready_1) begin
        _dataflow_counter_valid_1 <= 0;
      end 
      if((_dataflow_counter_ready_1 || !_dataflow_counter_valid_1) && 1) begin
        _dataflow_counter_valid_1 <= 1;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
    end else begin
      case(fsm)
        fsm_init: begin
          fsm <= fsm_1;
        end
        fsm_1: begin
          if(_tmp_1) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          fsm <= fsm_3;
        end
        fsm_3: begin
          fsm <= fsm_4;
        end
        fsm_4: begin
          if(_tmp_9) begin
            fsm <= fsm_5;
          end 
        end
      endcase
    end
  end

  localparam _tmp_fsm_0_1 = 1;
  localparam _tmp_fsm_0_2 = 2;
  localparam _tmp_fsm_0_3 = 3;
  localparam _tmp_fsm_0_4 = 4;
  localparam _tmp_fsm_0_5 = 5;
  localparam _tmp_fsm_0_6 = 6;
  localparam _tmp_fsm_0_7 = 7;
  localparam _tmp_fsm_0_8 = 8;
  localparam _tmp_fsm_0_9 = 9;
  localparam _tmp_fsm_0_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_fsm_0 <= _tmp_fsm_0_init;
      _d1__tmp_fsm_0 <= _tmp_fsm_0_init;
      _d2__tmp_fsm_0 <= _tmp_fsm_0_init;
      myram_1_addr <= 0;
      _tmp_16 <= 0;
      _tmp_8 <= 0;
      __tmp_fsm_0_cond_1_0_1 <= 0;
      __tmp_fsm_0_cond_1_0_2 <= 0;
      _tmp_12 <= 0;
      __tmp_fsm_0_cond_2_1_1 <= 0;
      __tmp_fsm_0_cond_2_1_2 <= 0;
      _tmp_13 <= 0;
      _tmp_10 <= 0;
      _tmp_11 <= 0;
      _tmp_15 <= 0;
      __tmp_fsm_0_cond_5_2_1 <= 0;
      __tmp_fsm_0_cond_6_3_1 <= 0;
      __tmp_fsm_0_cond_6_3_2 <= 0;
      __tmp_fsm_0_cond_7_4_1 <= 0;
      __tmp_fsm_0_cond_7_4_2 <= 0;
      __tmp_fsm_0_cond_10_5_1 <= 0;
    end else begin
      _d1__tmp_fsm_0 <= _tmp_fsm_0;
      _d2__tmp_fsm_0 <= _d1__tmp_fsm_0;
      case(_d2__tmp_fsm_0)
        _tmp_fsm_0_1: begin
          if(__tmp_fsm_0_cond_1_0_2) begin
            _tmp_12 <= myram_1_rdata;
          end 
        end
        _tmp_fsm_0_2: begin
          if(__tmp_fsm_0_cond_2_1_2) begin
            _tmp_13 <= myram_1_rdata;
          end 
        end
        _tmp_fsm_0_6: begin
          if(__tmp_fsm_0_cond_6_3_2) begin
            _tmp_12 <= myram_1_rdata;
          end 
        end
        _tmp_fsm_0_7: begin
          if(__tmp_fsm_0_cond_7_4_2) begin
            _tmp_13 <= myram_1_rdata;
          end 
        end
      endcase
      case(_d1__tmp_fsm_0)
        _tmp_fsm_0_1: begin
          __tmp_fsm_0_cond_1_0_2 <= __tmp_fsm_0_cond_1_0_1;
        end
        _tmp_fsm_0_2: begin
          __tmp_fsm_0_cond_2_1_2 <= __tmp_fsm_0_cond_2_1_1;
        end
        _tmp_fsm_0_5: begin
          if(__tmp_fsm_0_cond_5_2_1) begin
            _tmp_15 <= 0;
          end 
        end
        _tmp_fsm_0_6: begin
          __tmp_fsm_0_cond_6_3_2 <= __tmp_fsm_0_cond_6_3_1;
        end
        _tmp_fsm_0_7: begin
          __tmp_fsm_0_cond_7_4_2 <= __tmp_fsm_0_cond_7_4_1;
        end
        _tmp_fsm_0_10: begin
          if(__tmp_fsm_0_cond_10_5_1) begin
            _tmp_15 <= 0;
          end 
        end
      endcase
      case(_tmp_fsm_0)
        _tmp_fsm_0_init: begin
          if(fsm == 3) begin
            myram_1_addr <= -1;
            _tmp_16 <= 0;
            _tmp_8 <= 32;
          end 
          if((fsm == 3) && 1) begin
            _tmp_fsm_0 <= _tmp_fsm_0_1;
          end 
        end
        _tmp_fsm_0_1: begin
          myram_1_addr <= myram_1_addr + 1;
          _tmp_8 <= (_tmp_8 > 0)? _tmp_8 - 1 : _tmp_8;
          __tmp_fsm_0_cond_1_0_1 <= 1;
          _tmp_fsm_0 <= _tmp_fsm_0_2;
        end
        _tmp_fsm_0_2: begin
          myram_1_addr <= myram_1_addr + 1;
          _tmp_8 <= (_tmp_8 > 0)? _tmp_8 - 1 : _tmp_8;
          __tmp_fsm_0_cond_2_1_1 <= 1;
          _tmp_fsm_0 <= _tmp_fsm_0_3;
        end
        _tmp_fsm_0_3: begin
          _tmp_fsm_0 <= _tmp_fsm_0_4;
        end
        _tmp_fsm_0_4: begin
          _tmp_fsm_0 <= _tmp_fsm_0_5;
        end
        _tmp_fsm_0_5: begin
          _tmp_10 <= _tmp_12;
          _tmp_11 <= _tmp_13;
          _tmp_15 <= 1;
          _tmp_16 <= _tmp_8 == 0;
          __tmp_fsm_0_cond_5_2_1 <= 1;
          _tmp_fsm_0 <= _tmp_fsm_0_6;
        end
        _tmp_fsm_0_6: begin
          myram_1_addr <= myram_1_addr + 1;
          _tmp_8 <= (_tmp_8 > 0)? _tmp_8 - 1 : _tmp_8;
          __tmp_fsm_0_cond_6_3_1 <= 1;
          _tmp_fsm_0 <= _tmp_fsm_0_7;
        end
        _tmp_fsm_0_7: begin
          myram_1_addr <= myram_1_addr + 1;
          _tmp_8 <= (_tmp_8 > 0)? _tmp_8 - 1 : _tmp_8;
          __tmp_fsm_0_cond_7_4_1 <= 1;
          _tmp_fsm_0 <= _tmp_fsm_0_8;
        end
        _tmp_fsm_0_8: begin
          _tmp_fsm_0 <= _tmp_fsm_0_9;
        end
        _tmp_fsm_0_9: begin
          _tmp_fsm_0 <= _tmp_fsm_0_10;
        end
        _tmp_fsm_0_10: begin
          if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 0)) begin
            _tmp_10 <= _tmp_12;
          end 
          if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 0)) begin
            _tmp_11 <= _tmp_13;
          end 
          if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 0)) begin
            _tmp_15 <= !_tmp_16;
            _tmp_16 <= _tmp_8 == 0;
          end 
          __tmp_fsm_0_cond_10_5_1 <= 1;
          if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 0) && (_tmp_8 == 0)) begin
            _tmp_fsm_0 <= _tmp_fsm_0_init;
          end 
          if((_tmp_5 || !_tmp_2) && (_tmp_6 || !_tmp_3) && (_tmp_7 || !_tmp_4) && (_tmp_14 == 0) && (_tmp_8 > 0)) begin
            _tmp_fsm_0 <= _tmp_fsm_0_6;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      sum0 <= 0;
      sum1 <= 0;
      _seq_cond_0_1 <= 0;
    end else begin
      if(_seq_cond_0_1) begin
        $display("sum=%d expected_sum=%d", (sum0 + sum1), 1984);
      end 
      if(_dataflow__variable_ovalid_4) begin
        sum0 <= sum0 + _dataflow__variable_odata_4;
        $display("rdata0_data=%d", _dataflow__variable_odata_4);
      end 
      if(_dataflow__variable_ovalid_5) begin
        sum1 <= sum1 + _dataflow__variable_odata_5;
        $display("rdata1_data=%d", _dataflow__variable_odata_5);
      end 
      _seq_cond_0_1 <= _dataflow__variable_ovalid_5 && (_dataflow__variable_odata_3 == 1);
    end
  end


endmodule



module myram
(
  input CLK,
  input [14-1:0] myram_0_addr,
  output [32-1:0] myram_0_rdata,
  input [32-1:0] myram_0_wdata,
  input myram_0_wenable,
  input [14-1:0] myram_1_addr,
  output [32-1:0] myram_1_rdata,
  input [32-1:0] myram_1_wdata,
  input myram_1_wenable
);

  reg [14-1:0] myram_0_daddr;
  reg [14-1:0] myram_1_daddr;
  reg [32-1:0] mem [0:16384-1];

  always @(posedge CLK) begin
    if(myram_0_wenable) begin
      mem[myram_0_addr] <= myram_0_wdata;
    end 
    myram_0_daddr <= myram_0_addr;
  end

  assign myram_0_rdata = mem[myram_0_daddr];

  always @(posedge CLK) begin
    if(myram_1_wenable) begin
      mem[myram_1_addr] <= myram_1_wdata;
    end 
    myram_1_daddr <= myram_1_addr;
  end

  assign myram_1_rdata = mem[myram_1_daddr];

endmodule
"""


def test():
    veriloggen.reset()
    test_module = thread_ram_read_dataflow_reuse.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)
